import inspect
import logging
import sys
from itertools import chain
from attrs import define
from datamaestro_text.interfaces.trec import write_run_dict
import pandas as pd
from pathlib import Path
from typing import DefaultDict, Dict, List, Protocol, Union, Tuple, Optional
import ir_measures
from experimaestro import Task, Param, pathgenerator, Annotated, tags, TagDict
from datamaestro_text.data.ir import (
    Adhoc,
    AdhocAssessments,
    AdhocRun,
    AdhocRunDict,
    Documents,
    AdhocResults,
    IDItem,
)
from datamaestro_text.data.ir.trec import TrecAdhocRun, TrecAdhocResults
from datamaestro_text.transforms.ir import TopicWrapper
from xpmir.measures import Measure
import xpmir.measures as m
from xpmir.metrics import evaluator
from xpmir.rankers import Retriever
from xpmir.utils.logging import easylog
from experimaestro.launchers import Launcher

logger = easylog()


def get_evaluator(metrics: List[ir_measures.Metric], assessments: AdhocAssessments):
    qrels = {
        assessedTopic.topic_id: {r.doc_id: r.rel for r in assessedTopic.assessments}
        for assessedTopic in assessments.iter()
    }
    return evaluator(metrics, qrels)


class BaseEvaluation(Task):
    """Base class for evaluation tasks"""

    measures: Param[List[Measure]] = [m.AP, m.P @ 20, m.nDCG, m.nDCG @ 20, m.RR]
    """List of metrics"""

    aggregated: Annotated[Path, pathgenerator("aggregated.txt")]
    """Path for aggregated results"""

    detailed: Annotated[Path, pathgenerator("detailed.dat")]
    """Path for detailed results"""

    with_run: Param[bool] = False
    """Saves the run together with the evaluation"""

    run_path: Annotated[Path, pathgenerator("run.txt")]
    """Path to save the run (TREC format). Only used if with_run is True"""

    def task_outputs(self, dep):
        results = dep(
            TrecAdhocResults.C(
                id="",
                results=self.aggregated,
                detailed=self.detailed,
                metrics=self.measures,
            )
        )
        if self.with_run:
            return results, dep(TrecAdhocRun.C(id="", path=self.run_path))
        return results

    def _execute(self, run: AdhocRunDict, assessments):
        """Evaluate an IR ad-hoc run with trec-eval"""

        if self.with_run:
            logging.info("Writing the run")
            write_run_dict(run, self.run_path)

        evaluator = get_evaluator([measure() for measure in self.measures], assessments)

        def print_line(fp, measure, scope, value):
            fp.write("{:25s} {:10s} {:.4f}\n".format(measure, scope, value))

        with self.detailed.open("w") as fp:
            for metric in evaluator.iter_calc(run):
                print_line(fp, str(metric.measure), metric.query_id, metric.value)

        with self.aggregated.open("w") as fp:
            for key, value in evaluator.calc_aggregate(run).items():
                print_line(fp, str(key), "all", value)


def get_run(retriever: Retriever, dataset: Adhoc) -> AdhocRunDict:
    """Returns the scored documents for each topic in a dataset"""
    results = retriever.retrieve_all(
        {topic[IDItem].id: topic for topic in dataset.topics.iter()}
    )
    return {
        qid: {sd.document[IDItem].id: sd.score for sd in scoredocs}
        for qid, scoredocs in results.items()
    }


def evaluate(retriever: Retriever, dataset: Adhoc, measures: List[str], details=False):
    """Evaluate a retriever on a given dataset

    :param retriever: The retriever to evaluate
    :param dataset: The dataset on which to evaluate
    :param measures: The list of measures to compute (using ir_measures)
    :param details: if query-level metrics should be reported, defaults to False
    :return: The metrics (if details is False) or a tuple (metrics, detailed metrics)
    """
    evaluator = get_evaluator(
        [ir_measures.parse_measure(m) for m in measures], dataset.assessments
    )
    run = get_run(retriever, dataset)

    aggregators = {m: m.aggregator() for m in evaluator.measures}
    details = DefaultDict(lambda: {}) if details else None
    for metric in evaluator.iter_calc(run):
        aggregators[metric.measure].add(metric.value)
        if details is not None:
            details[str(metric.measure)][metric.query_id] = metric.value

    metrics = {str(m): agg.result() for m, agg in aggregators.items()}
    if details is not None:
        return metrics, details

    return metrics


class RunEvaluation(BaseEvaluation, Task):
    """Evaluate a run"""

    run: Param[TrecAdhocRun]
    assessments: Param[AdhocAssessments]

    def execute(self):
        run = ir_measures.read_trec_run(self.run.path, self.assessments)
        return self._execute(run)


class Evaluate(BaseEvaluation, Task):
    """Evaluate a retriever directly (without generating the run explicitly)"""

    dataset: Param[Adhoc]
    """The dataset for retrieval"""

    retriever: Param[Retriever]
    """The retriever to evaluate"""

    topic_wrapper: Param[Optional[TopicWrapper]] = None
    """Topic extractor"""

    def execute(self):
        self.retriever.initialize()
        run = get_run(self.retriever, self.dataset)
        self._execute(run, self.dataset.assessments)


class RetrieverFactory(Protocol):
    """Generates a retriever for a given dataset"""

    def __call__(self, dataset: Documents) -> Retriever:
        ...


class Evaluations:
    """Holds experiment results for several models
    on one dataset"""

    dataset: Adhoc
    measures: List[Measure]
    results: List[BaseEvaluation]
    per_tag: Dict[TagDict, AdhocResults]

    topic_wrapper: Optional[TopicWrapper]

    def __init__(
        self,
        dataset: Adhoc,
        measures: List[Measure],
        *,
        topic_wrapper: Optional[TopicWrapper] = None,
    ):
        self.dataset = dataset
        self.measures = measures
        self.results = []
        self.per_tags = {}
        self.topic_wrapper = topic_wrapper

    def evaluate_retriever(
        self,
        key: str,
        retriever: Union[Retriever, RetrieverFactory],
        launcher: Launcher = None,
        *,
        init_tasks=[],
        with_run=False,
    ) -> "EvaluationResult":
        """Evaluates a retriever

        :param key: test collection key
        :param retriever: the retriever (or the retriever factory)
        """
        if not isinstance(retriever, Retriever):
            kwargs = {}
            sig = inspect.signature(retriever)
            kw_only = set(
                name
                for name, p in sig.parameters.items()
                if p.kind == inspect.Parameter.KEYWORD_ONLY
            )
            if "key" in kw_only:
                kwargs["key"] = key
            retriever = retriever(self.dataset.documents, **kwargs)

        task = Evaluate.C(
            retriever=retriever,
            measures=self.measures,
            dataset=self.dataset,
            topic_wrapper=self.topic_wrapper,
            with_run=with_run,
        )

        evaluation = task.submit(launcher=launcher, init_tasks=init_tasks)

        run = None
        if with_run:
            evaluation, run = evaluation

        self.add(evaluation)

        # Use retriever tags
        retriever_tags = tags(evaluation)
        if retriever_tags:
            self.per_tags[retriever_tags] = evaluation

        return EvaluationResult(key, evaluation, run, task)

    def add(self, *results: BaseEvaluation):
        self.results.extend(results)

    def output_results_per_tag(self, file=sys.stdout):
        return self.to_dataframe().to_markdown(file)

    def to_dataframe(self) -> pd.DataFrame:
        # Get all the tags
        tags = list(
            set(chain(*[tags_dict.keys() for tags_dict in self.per_tags.keys()]))
        )
        tags.sort()

        assert (
            len(tags) > 0
        ), "No tags found, please tag your evaluations to convert results to dataframe"

        # Get all the results and metrics
        to_process = []
        metrics = set()
        for tags_dict, evaluate in self.per_tags.items():
            try:
                results = evaluate.get_results()
                metrics.update(results.keys())
                to_process.append((tags_dict, results))
            except FileNotFoundError:
                logger.error("Cannot retrieve evaluation results for %s", tags_dict)

        # Sort metrics
        metrics = list(metrics)
        metrics.sort()

        # Table header
        columns = []
        for tag in tags:
            columns.append(["tag", tag])
        for metric in metrics:
            columns.append(["metric", metric])

        # Output the results
        rows = []
        for tags_dict, results in to_process:
            row = []
            # tag values
            for k in tags:
                row.append(str(tags_dict.get(k, "")))

            # metric values
            for metric in metrics:
                row.append(results.get(metric, ""))
            rows.append(row)

        index = pd.MultiIndex.from_tuples(columns)
        return pd.DataFrame(rows, columns=index)


@define
class EvaluationResult:
    key: str
    """Dataset identifier"""

    result: AdhocResults
    """Results"""

    run: Optional[AdhocRun]
    """The run (if available)"""

    task: Evaluate
    """The task for this result"""


class EvaluationsCollection:
    """A collection of evaluation

    This is useful to group all the evaluations to be conducted, and then
    to call the :py:meth:`evaluate_retriever`
    """

    collection: Dict[str, Evaluations]

    per_model: Dict[str, List[Tuple[str, AdhocResults]]]
    """List of results per model"""

    def __init__(self, **collection: Evaluations):
        self.collection = collection
        self.per_model = {}

    def evaluate_retriever(
        self,
        retriever: Union[Retriever, RetrieverFactory],
        launcher: Launcher = None,
        model_id: str = None,
        overwrite: bool = False,
        with_run: bool = False,
        init_tasks=[],
    ) -> list[EvaluationResult]:
        """Evaluate a retriever for all the evaluations in this collection (the
        tasks are submitted to the experimaestro scheduler)

        :param with_run: should the run be preserved (default False). Note that
            this changes the experiment ID.
        """
        if model_id is not None and not overwrite:
            assert (
                model_id not in self.per_model
            ), f"Model with ID `{model_id}` was already evaluated"

        results = []
        for key, evaluations in self.collection.items():
            result = evaluations.evaluate_retriever(
                key, retriever, launcher, init_tasks=init_tasks, with_run=with_run
            )
            results.append(result)

        # Adds to per model results
        if model_id is not None:
            self.per_model[model_id] = results

        return results

    def output_results(self, file=sys.stdout):
        """Print all the results"""
        for key, dsevaluations in self.collection.items():
            print(f"## Dataset {key}\n", file=file)  # noqa: T201
            for evaluation in dsevaluations.results:
                with evaluation.results.open("rt") as fp:
                    results = [f"- {line}" for line in fp.readlines()]
                    results = "".join(results)

                print(  # noqa: T201
                    f"### Results for {evaluation.__xpm__.tags()}" f"""\n{results}\n""",
                    file=file,
                )

    def to_dataframe(self) -> pd.DataFrame:
        """Returns a Pandas dataframe"""
        all_data = []
        for key, evaluations in self.collection.items():
            data = evaluations.to_dataframe()
            data["dataset"] = key
            all_data.append(data)
        return pd.concat(all_data, ignore_index=True)

    def output_results_per_tag(self, file=sys.stdout):
        """Outputs the results for each collection, based on the retriever tags
        to build the table
        """
        # Loop over all collections
        for key, evaluations in self.collection.items():
            print(f"## Dataset {key}\n", file=file)  # noqa: T201
            evaluations.output_results_per_tag(file)

    def output_model_results(self, model_id: str, file=sys.stdout):
        """Outputs the result of a model over various datasets (in markdown format)

        :param model_id: The model id, as given by :meth:`evaluate_retriever`
        :param file: The output stream, defaults to sys.stdout
        """
        all_results = {}
        all_metrics = set()
        for key, evaluation in self.per_model[model_id]:
            all_results[key] = evaluation.get_results()
            all_metrics.update(all_results[key].keys())
        all_metrics = sorted(all_metrics)

        file.write(f"| Dataset  | {' | '.join(all_metrics)}  |\n")  # noqa: E221
        file.write(f"|----| {'---|---'.join('' for _ in all_metrics)}---|\n")
        for key, values in all_results.items():
            file.write(f"| {key}")
            for metric in all_metrics:
                value = values.get(metric, "")
                file.write(f" | {value}")
            file.write(" |\n")



from typing import Any, List, Optional, Dict
import logging
import pandas as pd
from pathlib import Path
import click
import io
from functools import cached_property

import docstring_parser
from experimaestro import RunMode, Config
from experimaestro.exceptions import HandledException
from xpmir.evaluation import EvaluationsCollection
from xpmir.models import XPMIRHFHub
from xpmir.papers.results import PaperResults
from xpmir.experiments.learning import LearningExperimentHelper
from xpmir.learning.optim import TensorboardService


class UploadToHub:
    def __init__(self, model_id: Optional[str], doc: docstring_parser.Docstring):
        self.model_id = model_id
        self.doc = doc

    def send_scorer(
        self,
        models: Dict[str, Config],
        *,
        evaluations: Optional[EvaluationsCollection] = None,
        tb_logs: Dict[str, Path],
        add_des: str = "",
    ):
        """Upload the scorer(s) to the HuggingFace Hub

        :param models: The models to upload, each with a key
        :param tb_logs: The tensorboard logs
        :param evaluations: Models evaluations, defaults to None
        :param add_des: Extra documentation, defaults to ""
        """
        if self.model_id is None:
            return

        out = io.StringIO()
        out.write(
            """---
library_name: xpmir
---
"""
        )

        out.write(f"{self.doc}\n\n")
        out.write(f"{add_des}\n\n")

        out.write("\n## Using the model")
        out.write(
            f"""
The model can be loaded with [experimaestro
IR](https://experimaestro-ir.readthedocs.io/en/latest/)

If you want to use the model in further experiments with XPMIR,
use this code:
```py
from xpmir.models import AutoModel
from xpmir.models import AutoModel

model, init_tasks = AutoModel.load_from_hf_hub("{self.model_id}")
```


Use this code if you want to use the model in inference only:

```py
from xpmir.models import AutoModel
from xpmir.models import AutoModel

model = AutoModel.load_from_hf_hub("{self.model_id}", as_instance=True)
model.rsv("walgreens store sales average", "The average Walgreens salary ranges...")
```
"""
        )

        assert len(models) == 1, "Cannot deal with more than one variant"
        ((key, model),) = list(models.items())

        if evaluations is not None:
            out.write("\n## Results\n\n")
            evaluations.output_model_results(key, file=out)

        readme_md = out.getvalue()

        logging.info("Uploading to HuggingFace Hub")
        XPMIRHFHub(model, readme=readme_md, tb_logs=tb_logs).push_to_hub(
            repo_id=self.model_id, config={}
        )


class IRExperimentHelper(LearningExperimentHelper):
    """Helper for IR experiments"""

    def run(self, extra_args: List[str], configuration: Any):
        @click.option("--upload-to-hub", type=str)
        @click.command()
        def cli(upload_to_hub: str):
            try:
                results = self.callable(self, configuration)
            except Exception as e:
                logging.exception("Error while running the experiment")
                raise HandledException(e)
            self.xp.wait()

            if isinstance(results, PaperResults) and self.xp.run_mode == RunMode.NORMAL:
                if upload_to_hub is not None:
                    if configuration.title == "" and configuration.description == "":
                        doc = docstring_parser.parse(self.callable.__doc__)
                    else:
                        doc = f"# {configuration.title}\n{configuration.description}"
                    upload = UploadToHub(upload_to_hub, doc)

                    upload.send_scorer(
                        results.models,
                        evaluations=results.evaluations,
                        tb_logs=results.tb_logs,
                    )

                # Print the results
                df = results.evaluations.to_dataframe()
                pd.set_option("display.max_columns", None)
                pd.set_option("display.max_rows", None)
                pd.set_option("display.width", 200)
                print(df)  # noqa: T201

                # And save them

                csv_path = self.xp.resultspath / "results.csv"
                if not self.xp.resultspath.exists():
                    self.xp.resultspath.mkdir(parents=True, exist_ok=True)
                logging.info(f"Saved results in {csv_path.absolute()}")
                with csv_path.open("wt") as fp:
                    df.to_csv(fp, index=False)

        return cli(extra_args, standalone_mode=False)

    @cached_property
    def tensorboard_service(self):
        return self.xp.add_service(
            TensorboardService(self.xp, self.xp.resultspath / "runs")
        )


ir_experiment = IRExperimentHelper.decorator
"""Uses an IR experiment helper that provides

1. Tensorboard service (from Learning)
1. Upload to HuggingFace
1. Printing the evaluation results
"""



from functools import lru_cache, partial

try:
    from functools import cache
except ImportError:

    def cache(user_function, /):
        """Cache function for python 3.8"""
        return lru_cache(maxsize=None)(user_function)


def partial_cache(func, *args, **kwargs):
    """Combines a lru_cache with a partial"""
    return cache(partial(func, *args, **kwargs))



# This package contains all rankers

from abc import ABC, abstractmethod
from experimaestro import tqdm
from enum import Enum
from typing import (
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    TYPE_CHECKING,
)
import torch
import torch.nn as nn
import attrs
from experimaestro import Param, Config, Meta, field
from datamaestro_text.data.ir import (
    Documents,
    DocumentStore,
    create_record,
    IDItem,
)
from datamaestro_text.data.ir.base import DocumentRecord
from xpmir.utils.utils import Initializable
from xpmir.letor import Device, Random
from xpmir.learning import ModuleInitMode, ModuleInitOptions
from xpmir.learning.batchers import Batcher
from xpmir.learning.context import TrainerContext
from xpmir.learning.optim import Module
from xpmir.letor.records import (
    TopicRecord,
    BaseRecords,
    PairwiseRecord,
    PairwiseRecords,
    ProductRecords,
)
from xpmir.utils.utils import EasyLogger, easylog

if TYPE_CHECKING:
    from xpmir.evaluation import RetrieverFactory

logger = easylog()


@attrs.define()
class ScoredDocument:
    """A data structure that associated a score with a document"""

    document: DocumentRecord
    """The document"""

    score: float
    """The associated score"""

    def __repr__(self):
        return f"document({self.document}, {self.score})"

    def __lt__(self, other):
        return self.score < other.score


class ScorerOutputType(Enum):
    REAL = 0
    """An unbounded scalar value"""

    LOG_PROBABILITY = 1
    """A log probability, bounded by 0"""

    PROBABILITY = 2
    """A probability, in ]0,1["""


class Scorer(Config, Initializable, EasyLogger, ABC):
    """Query-document scorer

    A model able to give a score to a list of documents given a query
    """

    outputType: ScorerOutputType = ScorerOutputType.REAL
    """Determines the type of output scalar (log probability, probability, logit) """

    def __initialize__(self, options: ModuleInitOptions):
        """Initialize the scorer

        :param options: Options for initialization
        """
        pass

    def rsv(
        self,
        topic: Union[str, TopicRecord],
        documents: Union[List[ScoredDocument], ScoredDocument, str, List[str]],
    ) -> List[ScoredDocument]:
        # Convert into document records
        if isinstance(documents, str):
            documents = [ScoredDocument(create_record(text=documents), None)]
        elif isinstance(documents[0], str):
            documents = [
                ScoredDocument(create_record(text=scored_document), None)
                for scored_document in documents
            ]

        # Convert into topic record
        if isinstance(topic, str):
            topic = create_record(text=topic)

        return self.compute(topic, documents)

    @abstractmethod
    def compute(
        self, topic: TopicRecord, documents: Iterable[ScoredDocument]
    ) -> List[ScoredDocument]:
        """Score all documents with respect to the topic"""
        ...

    def eval(self):
        """Put the model in inference/evaluation mode"""
        pass

    def to(self, device):
        """Move the scorer to another device"""
        pass

    def getRetriever(
        self,
        retriever: "Retriever",
        batch_size: int,
        batcher: Batcher = Batcher.C(),
        top_k=None,
        device=None,
    ):
        """Returns a two stage re-ranker from this retriever and a scorer

        :param device: Device for the ranker or None if no change should be made

        :param batch_size: The number of documents in each batch

        :param top_k: Number of documents to re-rank (or None for all)
        """
        return TwoStageRetriever.C(
            retriever=retriever,
            scorer=self,
            batchsize=batch_size,
            batcher=batcher,
            device=device,
            top_k=top_k if top_k else None,
        )


def scorer_retriever(
    documents: Documents,
    *,
    retrievers: "RetrieverFactory",
    scorer: Scorer,
    **kwargs,
):
    """Helper function that returns a two stage retriever. This is useful
    when used with partial (when the scorer is not known).

    :param documents: The document collection
    :param retrievers: A retriever factory
    :param scorer: The scorer
    :return: A retriever, calling the :meth:scorer.getRetriever
    """
    assert retrievers is not None, "The retrievers have not been given"
    assert scorer is not None, "The scorer has not been given"
    return scorer.getRetriever(retrievers(documents), **kwargs)


class RandomScorer(Scorer):
    """A random scorer"""

    random: Param[Random]
    """The random number generator"""

    def compute(
        self, record: TopicRecord, scored_documents: Iterable[ScoredDocument]
    ) -> List[ScoredDocument]:
        result = []
        random = self.random.state
        for scored_document in scored_documents:
            result.append(ScoredDocument(scored_document.document, random.random()))
        return result


class AbstractModuleScorerCall(Protocol):
    def __call__(self, inputs: "BaseRecords", info: Optional[TrainerContext]):
        ...


class AbstractModuleScorer(Scorer, Module):
    """Base class for all learnable scorer

    This class provides a `compute` method that calls the forward method,

    """

    # Ensures basic operations are redirected to torch.nn.Module methods
    __call__: AbstractModuleScorerCall = nn.Module.__call__
    to = nn.Module.to
    train = nn.Module.train

    def __init__(self):
        self.logger.info("Initializing %s", self)
        nn.Module.__init__(self)
        super().__init__()
        self._initialized = False

    def __str__(self):
        return f"scorer {self.__class__.__qualname__}"

    def eval(self):
        """Put the model in training mode"""
        self.train(False)

    def __initialize__(self, options: ModuleInitOptions):
        """Initialize a learnable scorer

        Initialization can either be determined by a checkpoint (if set) or
        otherwise (random or pre-trained checkpoint depending on the models)
        """
        # Sets the current random seed
        if options.random is not None:
            seed = options.random.randint((2**32) - 1)
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)

        return self

    def compute(
        self, topic: TopicRecord, scored_documents: Iterable[ScoredDocument]
    ) -> List[ScoredDocument]:

        # Prepare the inputs and call the model
        inputs = ProductRecords()
        inputs.add_topics(topic)

        inputs.add_documents(*[sd.document for sd in scored_documents])

        with torch.no_grad():
            scores = self(inputs, None).cpu().numpy()

        # Returns the scored documents
        scoredDocuments = []
        for i in range(len(scored_documents)):
            scoredDocuments.append(
                ScoredDocument(
                    scored_documents[i].document,
                    float(scores[i]),
                )
            )

        return scoredDocuments


class LearnableScorer(AbstractModuleScorer):
    """Learnable scorer

    A scorer with parameters that can be learnt"""

    def forward(self, inputs: "BaseRecords", info: Optional[TrainerContext]):
        """Computes the score of all (query, document) pairs

        Different subclasses can process the input more or
        less efficiently based on the `BaseRecords` instance (pointwise,
        pairwise, or structured)
        """
        raise NotImplementedError(f"forward in {self.__class__}")


class DuoLearnableScorer(LearnableScorer):
    """Base class for models that can score a triplet (query, document 1, document 2)"""

    def forward(self, inputs: "PairwiseRecords", info: Optional[TrainerContext]):
        """Returns scores for pairs of documents (given a query)"""
        raise NotImplementedError(f"abstract __call__ in {self.__class__}")


class Retriever(Config, ABC):
    """A retriever is a model to return top-scored documents given a query"""

    store: Param[Optional[DocumentStore]] = None
    """Give the document store associated with this retriever"""

    def initialize(self):
        pass

    def collection(self):
        """Returns the document collection object"""
        raise NotImplementedError()

    def retrieve_all(
        self, queries: Dict[str, TopicRecord]
    ) -> Dict[str, List[ScoredDocument]]:
        """Retrieves for a set of documents

        By default, iterate using `self.retrieve`, but this leaves some room open
        for optimization

        Args:

            queries: A dictionary where the key is the ID of the query, and the value
                is the text
        """
        results = {}
        for key, record in tqdm(list(queries.items())):
            results[key] = self.retrieve(record)
        return results

    @abstractmethod
    def retrieve(self, record: TopicRecord) -> List[ScoredDocument]:
        """Retrieves documents, returning a list sorted by decreasing score

        if `content` is true, includes the document full text
        """
        ...

    def _store(self) -> Optional[DocumentStore]:
        """Returns the associated document store (if any) that can be
        used to get the full text of the documents"""

    def get_store(self) -> Optional[DocumentStore]:
        return self.store or self._store()


class AbstractTwoStageRetriever(Retriever):
    """Abstract class for all two stage retrievers (i.e. scorers and duo-scorers)"""

    retriever: Param[Retriever]
    """The base retriever"""

    scorer: Param[Scorer]
    """The scorer used to re-rank the documents"""

    top_k: Param[Optional[int]]
    """The number of returned documents (if None, returns all the documents)"""

    batchsize: Meta[int] = 0
    """The batch size for the re-ranker"""

    batcher: Meta[Batcher] = field(default_factory=Batcher.C)
    """How to provide batches of documents"""

    device: Meta[Optional[Device]]
    """Device on which the model is run"""

    def initialize(self):
        self.retriever.initialize()
        self._batcher = self.batcher.initialize(self.batchsize)
        self.scorer.initialize(ModuleInitMode.DEFAULT.to_options())

        # Compute with the scorer
        if self.device is not None:
            self.scorer.to(self.device.value)


class TwoStageRetriever(AbstractTwoStageRetriever):
    """Use on retriever to select the top-K documents which are the re-ranked
    given a scorer"""

    def _retrieve(
        self,
        batch: List[ScoredDocument],
        query: str,
        scoredDocuments: List[ScoredDocument],
    ):
        scoredDocuments.extend(self.scorer.rsv(query, batch))

    def retrieve(self, record: TopicRecord):
        # Calls the retriever
        scoredDocuments = self.retriever.retrieve(record)

        # Scorer in evaluation mode
        self.scorer.eval()

        _scoredDocuments = []
        scoredDocuments = self._batcher.process(
            scoredDocuments, self._retrieve, record, _scoredDocuments
        )

        _scoredDocuments.sort(reverse=True)
        return _scoredDocuments[: (self.top_k or len(_scoredDocuments))]


class DuoTwoStageRetriever(AbstractTwoStageRetriever):
    """The two stage retriever for pairwise scorers.

    For pairwise scorer, we need to aggregate the pairwise scores in some
    way.
    """

    def _retrieve(
        self,
        batch: List[Tuple[ScoredDocument, ScoredDocument]],
        query: str,
        scoredDocuments: List[float],
    ):
        """call the function rsv to get the information for each batch
        because of the batchsize is independent on k, we may seperate the
        triplets belongs to the same query into different batches.
        """
        scoredDocuments.extend(self.rsv(query, batch))

    def retrieve(self, query: TopicRecord):
        """call the _retrieve function by using the batcher and do an
        aggregation of all the scores
        """
        # get the documents from the retriever
        scoredDocuments_previous = self.retriever.retrieve(query)

        # transform them into the pairs (i, j)
        # for i != j ranging from 1 to nb of documents
        pairs = []
        for i in range(len(scoredDocuments_previous)):
            for j in range(len(scoredDocuments_previous)):
                if i != j:
                    pairs.append(
                        (scoredDocuments_previous[i], scoredDocuments_previous[j])
                    )

        # Scorer in evaluation mode
        self.scorer.eval()

        _scores_pairs = []  # the scores for each pair of documents
        self._batcher.process(pairs, self._retrieve, query, _scores_pairs)

        # Use the sum aggregation strategy
        _scores_pairs = torch.Tensor(_scores_pairs).reshape(
            len(scoredDocuments_previous), -1
        )
        _scores_per_document = torch.sum(
            _scores_pairs, dim=1
        )  # scores for each document.

        # construct the ScoredDocument object from the score we just get.
        scoredDocuments = []
        for i in range(len(scoredDocuments_previous)):
            scoredDocuments.append(
                ScoredDocument(
                    scoredDocuments_previous[i], float(_scores_per_document[i])
                )
            )
        scoredDocuments.sort(reverse=True)
        return scoredDocuments[: (self.top_k or len(scoredDocuments))]

    def rsv(
        self,
        record: TopicRecord,
        documents: List[Tuple[ScoredDocument, ScoredDocument]],
    ) -> List[float]:
        """Given the query and documents in tuple
        return the score for each triplets
        """
        inputs = PairwiseRecords()
        for doc1, doc2 in documents:
            inputs.add(PairwiseRecord(record, doc1, doc2))

        with torch.no_grad():
            scores = self.scorer(inputs, None).cpu().float()  # shape (batchsizes)
            return scores.tolist()


ARGS = TypeVar("ARGS")
KWARGS = TypeVar("KWARGS")
T = TypeVar("T")


class DocumentsFunction(Protocol, Generic[KWARGS, ARGS, T]):
    def __call__(self, documents: Documents, *args: ARGS, **kwargs: KWARGS) -> T:
        ...


def document_cache(fn: DocumentsFunction[KWARGS, ARGS, T]):
    """Decorator

    Allows to cache the result of a function that depends
    on the document dataset ID
    """
    retrievers = {}

    def _fn(*args: ARGS, **kwargs: KWARGS):
        def cached(documents: Documents) -> T:
            dataset_id = documents.__identifier__().all

            if dataset_id not in retrievers:
                retrievers[dataset_id] = fn(documents, *args, **kwargs)

            return retrievers[dataset_id]

        return cached

    return _fn


class RetrieverHydrator(Retriever):
    """Hydrate retrieved results with document text"""

    retriever: Param[Retriever]
    """The retriever to hydrate"""

    store: Param[DocumentStore]
    """The store for document texts"""

    def initialize(self):
        return self.retriever.initialize()

    def retrieve(self, record: TopicRecord) -> List[ScoredDocument]:
        return [
            ScoredDocument(self.store.document_ext(sd.document[IDItem].id), sd.score)
            for sd in self.retriever.retrieve(record)
        ]



from typing import Dict
import logging
from torch.utils.tensorboard.writer import SummaryWriter


class Metric:
    """Represents a generic metric"""

    key: str
    """The key for this metric (used for logging)"""

    count: int
    """Number of samples for this value"""

    def __init__(self, key, count):
        self.key = key
        self.count = count

    def merge(self, other: "Metric"):
        assert other.__class__ is self.__class__
        self._merge(other)

    def _merge(self, other: "Metric"):
        raise NotImplementedError(f"_merge in {self.__class__}")

    def report(self, epoch: int, writer: SummaryWriter, prefix: str):
        raise NotImplementedError(f"report in {self.__class__}")


class ScalarMetric(Metric):
    """Represents a scalar metric"""

    sum: float

    def __init__(self, key: str, value: float, count: int):
        super().__init__(key, count)
        self.sum = value * count

    def _merge(self, other: "ScalarMetric"):
        self.sum += other.sum
        self.count += other.count

    def compute(self):
        return self.sum / self.count

    def report(self, step: int, writer: SummaryWriter, prefix: str):
        if self.count == 0:
            logging.warning("Count is 0 when reporting metrics")
        writer.add_scalar(
            f"{prefix}/{self.key}",
            self.sum / self.count,
            step,
        )


class Metrics:
    """Utility class to accumulate a set of metrics over batches
    of (potentially) different sizes"""

    def __init__(self):
        self.metrics: Dict[str, Metric] = {}

    def add(self, metric: Metric):
        if metric.key in self.metrics:
            self.metrics[metric.key].merge(metric)
        else:
            self.metrics[metric.key] = metric

    def merge(self, other: "Metrics"):
        for value in other.metrics.values():
            self.add(value)

    def report(self, epoch, writer, prefix):
        for value in self.metrics.values():
            value.report(epoch, writer, prefix)



import os
import re
import logging
import torch.nn as nn
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

import torch

from experimaestro.compat import cached_property
from experimaestro import Param, Constant, deprecate
from xpmir.distributed import DistributableModel
from xpmir.learning.optim import ModuleInitMode, ModuleInitOptions
from xpmir.text.encoders import (
    Encoder,
    TokensEncoder,
    DualTextEncoder,
    TextEncoder,
    TripletTextEncoder,
    EncoderOutput,
    RepresentationOutput,
)
from xpmir.utils.utils import easylog
from xpmir.learning.context import TrainerContext, TrainState
from xpmir.learning.parameters import ParametersIterator
from .tokenizers import (  # noqa: F401
    HFTokenizer,
    HFTokenizerBase,
    HFTokenizerAdapter,
    HFStringTokenizer,
    HFListTokenizer,
)
from .encoders import (  # noqa: F401
    HFModel,
    HFTokensEncoder,
    HFCLSEncoder,
)
from .base import HFMaskedLanguageModel  # noqa: F401


try:
    from transformers import (
        AutoModel,
        AutoTokenizer,
        AutoConfig,
        DataCollatorForLanguageModeling,
        AutoModelForMaskedLM,
    )
except Exception:
    logging.error("Install huggingface transformers to use these configurations")
    raise

from xpmir.text import TokenizedTexts

logger = easylog()
logger.setLevel(logging.INFO)


class BaseTransformer(Encoder):
    """Base transformer class from Huggingface"""

    model_id: Param[str] = "bert-base-uncased"
    """Model ID from huggingface"""

    trainable: Param[bool]
    """Whether BERT parameters should be trained"""

    layer: Param[int] = 0
    """Layer to use (0 is the last, -1 to use them all)"""

    dropout: Param[Optional[float]] = 0
    """(deprecated) Define a dropout for all the layers"""

    CLS: int
    SEP: int

    @cached_property
    def tokenizer(self):
        return AutoTokenizer.from_pretrained(self.model_id, use_fast=True)

    @property
    def pad_tokenid(self) -> int:
        return self.tokenizer.pad_token_id

    @cached_property
    def config(self):
        return AutoConfig.from_pretrained(self.model_id)

    @property
    def automodel(self):
        return AutoModel

    def __initialize__(self, options: ModuleInitOptions):
        """Initialize the HuggingFace transformer

        Args:
            options: loader options
        """
        super().__initialize__(options)

        # Load the model configuration
        if self.dropout != 0:
            self.config.hidden_dropout_prob = self.dropout
            self.config.attention_probs_dropout_prob = self.dropout

        local_files_only = os.environ("HF_HUB_OFFLINE")
        if options.mode == ModuleInitMode.NONE or options.mode == ModuleInitMode.RANDOM:
            self.model = self.automodel.from_config(self.config)
        else:
            self.model = self.automodel.from_pretrained(
                self.model_id, config=self.config, local_files_only=local_files_only
            )

        # Loads the tokenizer
        self._tokenizer = AutoTokenizer.from_pretrained(
            self.model_id, use_fast=True, local_files_only=local_files_only
        )

        self.CLS = self.tokenizer.cls_token_id
        self.SEP = self.tokenizer.sep_token_id

        if self.trainable:
            self.model.train()
        else:
            self.model.eval()

    def parameters(self, recurse=True):
        if self.trainable:
            return super().parameters(recurse)
        return []

    def train(self, mode: bool = True):
        # We should not make this layer trainable unless asked
        if mode:
            if self.trainable:
                self.model.train(mode)
        else:
            self.model.train(mode)

    def tokenize(self, text):
        return self.tokenizer.tokenize(text)

    def tok2id(self, tok):
        return self.tokenizer.vocab[tok]

    def static(self):
        return not self.trainable

    def batch_tokenize(
        self,
        texts: Union[List[str], List[Tuple[str, str]]],
        batch_first=True,
        maxlen=None,
        mask=False,
    ) -> TokenizedTexts:
        if maxlen is None:
            maxlen = self.tokenizer.model_max_length
        else:
            maxlen = min(maxlen, self.tokenizer.model_max_length)

        assert batch_first, "Batch first is the only option"

        r = self.tokenizer(
            list(texts),
            max_length=maxlen,
            truncation=True,
            padding=True,
            return_tensors="pt",
            return_length=True,
            return_attention_mask=mask,
        )
        return TokenizedTexts(
            None,
            r["input_ids"].to(self.device),
            r["length"],
            r.get("attention_mask", None),
            r.get("token_type_ids", None),  # if r["token_type_ids"] else None
        )

    def id2tok(self, idx):
        if torch.is_tensor(idx):
            if len(idx.shape) == 0:
                return self.id2tok(idx.item())
            return [self.id2tok(x) for x in idx]
        # return self.tokenizer.ids_to_tokens[idx]
        return self.tokenizer.id_to_token(idx)

    def lexicon_size(self) -> int:
        return self.tokenizer._tokenizer.get_vocab_size()

    def maxtokens(self) -> int:
        try:
            return self.tokenizer.max_model_length
        except Exception:
            logging.warning(
                "No `max_model_length` in the tokenizer, "
                "defaulting to `model.config.max_position_embeddings` instead"
            )
            return self.config.max_position_embeddings

    def dim(self):
        return self.config.hidden_size

    @property
    def vocab_size(self) -> int:
        """Returns the size of the vocabulary"""
        return self.tokenizer.vocab_size


class TransformerTokensEncoder(BaseTransformer, TokensEncoder):
    """A tokens encoder based on HuggingFace"""

    def forward(self, toks: TokenizedTexts, all_outputs=False):
        outputs = self.model(
            toks.ids.to(self.device),
            attention_mask=toks.mask.to(self.device) if toks.mask is not None else None,
        )
        if all_outputs:
            return outputs
        return outputs.last_hidden_state


@deprecate
class TransformerVocab(TransformerTokensEncoder):
    """Old tokens encoder"""

    pass


class SentenceTransformerTextEncoder(TextEncoder):
    """A Sentence Transformers text encoder"""

    model_id: Param[str] = "sentence-transformers/all-MiniLM-L6-v2"

    def __initialize__(self, options: ModuleInitOptions):
        super().__initialize__(options)
        from sentence_transformers import SentenceTransformer

        self.model = SentenceTransformer(self.model_id)

    def forward(self, texts: List[str]) -> torch.Tensor:
        return self.model.encode(texts)


class OneHotHuggingFaceEncoder(TextEncoder):
    """A tokenizer which encodes the tokens into 0 and 1 vector
    1 represents the text contains the token and 0 otherwise"""

    model_id: Param[str] = "bert-base-uncased"
    """Model ID from huggingface"""

    maxlen: Param[Optional[int]] = None
    """Max length for texts"""

    version: Constant[int] = 2

    def __initialize__(self, options: ModuleInitOptions):
        super().__initialize__()
        self._tokenizer = AutoTokenizer.from_pretrained(self.model_id, use_fast=True)
        self.CLS = self._tokenizer.cls_token_id
        self.SEP = self._tokenizer.sep_token_id
        self.PAD = self._tokenizer.pad_token_id
        self._dummy_params = nn.Parameter(torch.Tensor())

    @property
    def device(self):
        return self._dummy_params.device

    @cached_property
    def tokenizer(self):
        return self._tokenizer

    def batch_tokenize(self, texts):
        r = self.tokenizer(
            list(texts),
            max_length=self.maxlen,
            truncation=True,
            padding=True,
            return_tensors="pt",
        )
        return r["input_ids"]

    def forward(self, texts: List[str]) -> torch.Tensor:
        """Returns a batch x vocab tensor"""
        tokenized_ids = self.batch_tokenize(texts)
        batch_size = len(texts)
        x = torch.zeros(batch_size, self.dimension)
        x[torch.arange(batch_size).unsqueeze(-1), tokenized_ids] = 1
        x[:, [self.PAD, self.SEP, self.CLS]] = 0
        return x.to(self.device)

    @property
    def dimension(self):
        return self.tokenizer.vocab_size

    def static(self):
        return False


@deprecate
class HuggingfaceTokenizer(OneHotHuggingFaceEncoder):
    """The old encoder for one hot"""

    pass


class TransformerEncoder(BaseTransformer, TextEncoder, DistributableModel):
    """Encodes using the [CLS] token"""

    maxlen: Param[Optional[int]] = None

    def forward(self, texts: List[str]):
        tokenized = self.batch_tokenize(
            texts,
            maxlen=self.maxlen if self.maxlen is not None else self.maxtokens(),
            mask=True,
        )

        with torch.set_grad_enabled(torch.is_grad_enabled() and self.trainable):
            y = self.model(tokenized.ids, attention_mask=tokenized.mask.to(self.device))

        # Assumes that [CLS] is the first token
        return y.last_hidden_state[:, 0]

    @property
    def dimension(self):
        return self.dim()

    def with_maxlength(self, maxlen: int):
        return TransformerTextEncoderAdapter(encoder=self, maxlen=maxlen)

    def distribute_models(self, update):
        self.model = update(self.model)


class TransformerTextEncoderAdapter(TextEncoder, DistributableModel):
    encoder: Param[TransformerEncoder]
    maxlen: Param[Optional[int]] = None

    def __initialize__(self, options: ModuleInitOptions):
        self.encoder.__initialize__(options)

    @property
    def dimension(self):
        return self.encoder.dimension

    def forward(self, texts: List[str], maxlen=None):
        return self.encoder.forward(texts, maxlen=self.maxlen)

    def static(self):
        return self.encoder.static()

    @property
    def vocab_size(self):
        return self.encoder.vocab_size

    def distribute_models(self, update):
        self.encoder.model = update(self.encoder.model)


class DualTransformerEncoder(BaseTransformer, DualTextEncoder):
    """Encodes the (query, document pair) using the [CLS] token

    maxlen: Maximum length of the query document pair (in tokens) or None if
    using the transformer limit
    """

    maxlen: Param[Optional[int]] = None

    version: Constant[int] = 2

    def forward(self, texts: List[Tuple[str, str]]) -> EncoderOutput:
        tokenized = self.batch_tokenize(
            texts,
            maxlen=self.maxlen if self.maxlen is not None else self.maxtokens(),
            mask=True,
        )

        with torch.set_grad_enabled(torch.is_grad_enabled() and self.trainable):
            kwargs = {}
            if tokenized.token_type_ids is not None:
                kwargs["token_type_ids"] = tokenized.token_type_ids.to(self.device)

            y = self.model(
                tokenized.ids, attention_mask=tokenized.mask.to(self.device), **kwargs
            )

        # Assumes that [CLS] is the first token
        return RepresentationOutput(y.last_hidden_state[:, 0])

    @property
    def dimension(self) -> int:
        return self.model.config.hidden_size


class DualDuoBertTransformerEncoder(BaseTransformer, TripletTextEncoder):
    """Vector encoding of a (query, document, document) triplet

    Be like: [cls] query [sep] doc1 [sep] doc2 [sep]

    """

    maxlen_query: Param[int] = 64
    """Maximum length for the query, the first document and the second one"""

    maxlen_doc: Param[int] = 224
    """Maximum length for the query, the first document and the second one"""

    def __initialize__(self, options: ModuleInitOptions):
        super().__initialize__(options)

        # Add an extra token type
        data = self.model.embeddings.token_type_embeddings.weight.data
        if len(data) < 3:
            logger.info("Adding an extra token type in transformer")
            data = torch.cat((data, torch.zeros(1, data.shape[1])))
            self.model.embeddings.token_type_embeddings = nn.Embedding.from_pretrained(
                data, freeze=False
            )

    def batch_tokenize(
        self,
        texts: List[Tuple[str, str, str]],
        batch_first=True,
        mask=False,
    ) -> TokenizedTexts:

        assert batch_first, "Batch first is the only option"

        query = self.tokenizer(
            [triplet[0] for triplet in texts],
            max_length=self.maxlen_query,
            truncation=True,
        )

        document_1 = self.tokenizer(
            [triplet[1] for triplet in texts],
            max_length=self.maxlen_doc,
            truncation=True,
        )

        document_2 = self.tokenizer(
            [triplet[2] for triplet in texts],
            max_length=self.maxlen_doc,
            truncation=True,
        )

        new_input_ids = []
        new_attention_mask = []
        new_token_type_ids = []
        length_factory = (
            []
        )  # [[query_length, document_1_length, document_2_length, total_length],..]
        new_length = []
        maxlen = 0
        batch_size = len(query["input_ids"])

        # calculate the maxlen of the sum for the 3 texts and stock them
        for index in range(batch_size):
            query_length = len(query["input_ids"][index])
            document_1_length = len(document_1["input_ids"][index]) - 1
            document_2_length = len(document_2["input_ids"][index]) - 1
            total_length_at_index = query_length + document_1_length + document_2_length
            if total_length_at_index > maxlen:
                maxlen = total_length_at_index
            length_factory.append(
                [
                    query_length,
                    document_1_length,
                    document_2_length,
                    total_length_at_index,
                ]
            )

        for index in range(batch_size):
            new_input_ids.append(
                query["input_ids"][index]
                + document_1["input_ids"][index][1:]
                + document_2["input_ids"][index][1:]
                + [self.pad_tokenid] * (maxlen - length_factory[index][3])
            )
            new_attention_mask.append(
                [1] * length_factory[index][3]
                + [0] * (maxlen - length_factory[index][3])
            )
            new_token_type_ids.append(
                [0] * length_factory[index][0]
                + [1] * length_factory[index][1]
                + [2] * length_factory[index][2]
                + [0] * (maxlen - length_factory[index][3])
            )
            new_length.append(length_factory[index][3])

        new_input_ids = torch.Tensor(new_input_ids).type(torch.long)
        new_attention_mask = torch.Tensor(new_attention_mask).type(torch.long)
        new_token_type_ids = torch.Tensor(new_token_type_ids).type(torch.long)
        new_length = torch.Tensor(new_length).type(torch.long)

        return TokenizedTexts(
            None,
            new_input_ids.to(self.device),
            new_length,
            new_attention_mask if mask else None,
            new_token_type_ids.to(self.device),
        )

    def forward(self, texts: List[Tuple[str, str, str]]):

        tokenized = self.batch_tokenize(texts, mask=True)

        with torch.set_grad_enabled(torch.is_grad_enabled() and self.trainable):
            y = self.model(
                tokenized.ids,
                token_type_ids=tokenized.token_type_ids,
                attention_mask=tokenized.mask.to(self.device),
            )

        # Assumes that [CLS] is the first token
        # shape of y.last_hidden_state: (1, len(texts), dimension)
        return y.last_hidden_state[:, 0]

    @property
    def dimension(self) -> int:
        return self.model.config.hidden_size

    # def distribute_models(self, update):
    #     self.model = update(self.model)


@dataclass
class MLMModelOutput:
    """Format for the output of the model during Masked Language Modeling"""

    logits: torch.LongTensor
    labels: torch.Tensor


class MLMEncoder(BaseTransformer, DistributableModel):
    """Implementation of the encoder for the Masked Language Modeling task"""

    maxlen: Param[Optional[int]] = None

    mlm_probability: Param[float] = 0.2
    """Probability to mask tokens"""

    datacollator: DataCollatorForLanguageModeling = None

    @property
    def automodel(self):
        return AutoModelForMaskedLM

    def __initialize__(self, options: ModuleInitOptions):
        super().__initialize__(options)
        self.datacollator = DataCollatorForLanguageModeling(
            tokenizer=self.tokenizer,
            mlm_probability=self.mlm_probability,
            return_tensors="pt",
        )

    def forward(self, texts: List[str], info: TrainerContext = None) -> MLMModelOutput:
        tokenized = self.batch_tokenize(texts, mask=True)

        masked = self.datacollator.torch_mask_tokens(tokenized.ids.cpu())

        with torch.set_grad_enabled(torch.is_grad_enabled() and self.trainable):
            y = self.model(
                input_ids=masked[0].to(self.device),
                labels=masked[1].to(self.device),
                attention_mask=tokenized.mask.to(self.device),
            )

        # Maybe easier to simply returns the object returned by the BertForMaskedLM?
        return MLMModelOutput(logits=y.logits, labels=masked[1])

    @property
    def dimension(self) -> int:
        return self.config.hidden_size

    def distribute_models(self, update):
        self.model = update(self.model)


class LayerSelector(ParametersIterator):
    """This class can be used to pick some of the transformer layers"""

    # For freezing everything except the embeddings
    re_layer: Param[str] = r"""(?:encoder|transformer)\.layer\.(\d+)\."""

    transformer: Param[BaseTransformer]
    """The model for which layers are selected"""

    pick_layers: Param[int] = 0
    """Counting from the first processing layers (can be negative, i.e. -1 meaning
    until the last layer excluded, etc. / 0 means no layer)"""

    select_embeddings: Param[bool] = False
    """Whether to pick the embeddings layer"""

    select_feed_forward: Param[bool] = False
    """Whether to pick the feed forward of Transformer layers"""

    def __post_init__(self):
        self._re_layer = re.compile(self.re_layer)

    def __validate__(self):
        if (
            not (self.select_embeddings or self.select_feed_forward)
            and self.pick_layers == 0
        ):
            raise AssertionError("The layer selector will select nothing")

    @cached_property
    def nlayers(self):
        count = 0
        for name, _ in self.transformer.model.named_parameters():
            if m := self._re_layer.search(name):
                count = max(count, int(m.group(1)))
        return count

    def should_pick(self, name: str) -> bool:
        if self.select_embeddings and ("embeddings." in name):
            return True

        if self.select_feed_forward and ("intermediate" in name):
            return True

        if self.pick_layers != 0:
            if m := self._re_layer.search(name):
                layer = int(m.group(1))
                if self.pick_layers < 0:
                    return layer <= self.nlayers + self.pick_layers
                return layer < self.pick_layers

        return False

    def iter(self):
        for name, params in self.transformer.model.named_parameters():
            yield f"model.{name}", params, self.should_pick(name)

    def after(self, state: TrainState):
        if not self._initialized:
            self._initialized = True
            for name, param in self.transformer.model.named_parameters():
                if self.should_freeze(name):
                    logger.info("Freezing layer %s", name)
                    param.requires_grad = False


class TransformerTokensEncoderWithMLMOutput(TransformerTokensEncoder):
    """Transformer that output logits over the vocabulary"""

    @property
    def automodel(self):
        return AutoModelForMaskedLM



import torch
import itertools
from datamaestro_text.data.ir import (
    TopicRecord,
    DocumentRecord,
    TextItem,
    create_record,
)
from typing import (
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)


class PointwiseRecord:
    """A record from a pointwise sampler"""

    # The query
    topic: TopicRecord

    # The document
    document: DocumentRecord

    # The relevance
    relevance: Optional[float]

    def __init__(
        self,
        topic: TopicRecord,
        document: DocumentRecord,
        relevance: Optional[float] = None,
    ):
        self.topic = topic
        self.document = document
        self.relevance = relevance

    @property
    def query(self):
        return self.topic


RT = TypeVar("RT")


class BaseRecords(List[RT]):
    """Base records just exposes iterables on (query, document) pairs

    Records can be structured, i.e. the same queries and documents
    can be used more than once. To allow optimization (e.g. pre-computing
    document/query representation),
    """

    topics: Iterable[TopicRecord]
    documents: Iterable[DocumentRecord]
    is_product = False

    @property
    def unique_topics(self) -> Iterable[TopicRecord]:
        return self.topics

    unique_queries = unique_topics

    @property
    def unique_documents(self) -> Iterable[DocumentRecord]:
        return self.documents

    @property
    def queries(self):
        """Deprecated: use topics"""
        return self.topics

    def pairs(self) -> Tuple[Iterable[int], Iterable[int]]:
        """Returns two iterators (over queries and documents)

        Returns the list of query/document indices for which we should compute
        the score, or None if all (cartesian product). This method should be
        used with `unique` set to true to get the queries/documents
        """
        raise NotImplementedError(f"pairs() in {self.__class__}")

    def __getitem__(self, ix: Union[slice, int]):
        """Sub-sample"""
        raise NotImplementedError(f"__getitem__() in {self.__class__}")

    def __len__(self):
        """Returns the number of records

        The length is dependant on the type of records, and is mainly used
        to divide the data into batches
        """
        raise NotImplementedError(f"__len__() in {self.__class__}")


class PointwiseRecords(BaseRecords[PointwiseRecord]):
    """Pointwise records are a set of triples (query, document, relevance)"""

    # The queries
    topics: List[TopicRecord]

    # Text of the documents
    documents: List[DocumentRecord]

    # The relevances
    relevances: List[float]

    def __init__(self):
        self.topics = []
        self.documents = []
        self.relevances = []

    def add(self, record: PointwiseRecord):
        self.queries.append(record.query)
        self.relevances.append(record.relevance or 0)
        self.documents.append(record.document)

    def __len__(self):
        return len(self.queries)

    def __getitem__(self, ix: Union[slice, int]):
        if isinstance(ix, slice):
            records = PointwiseRecords()
            for i in range(ix.start, min(ix.stop, len(self.topics)), ix.step or 1):
                records.add(
                    PointwiseRecord(
                        self.topics[i], self.documents[i], self.relevances[i]
                    )
                )
            return records

        return PointwiseRecord(self.topics[ix], self.documents[ix], self.relevances[ix])

    def pairs(self) -> Tuple[List[int], List[int]]:
        ix = list(range(len(self.queries)))
        return (ix, ix)

    @staticmethod
    def from_texts(
        topics: List[str],
        documents: List[str],
        relevances: Optional[List[float]] = None,
    ):
        records = PointwiseRecords()
        records.topics = list(map(lambda t: create_record(text=t), topics))
        records.documents = list(map(lambda t: create_record(text=t), documents))
        records.relevances = relevances
        return records


class PairwiseRecord:
    """A pairwise record is composed of a query, a positive and a negative document"""

    query: TopicRecord
    positive: DocumentRecord
    negative: DocumentRecord

    def __init__(
        self, query: TopicRecord, positive: DocumentRecord, negative: DocumentRecord
    ):
        self.query = query
        self.positive = positive
        self.negative = negative


class PairwiseRecordWithTarget(PairwiseRecord):
    """A pairwise record is composed of a query, a positive and a negative
    document, and the indetifier which says the one on the first is pos or
    neg"""

    target: int

    def __init__(
        self,
        query: TopicRecord,
        positive: DocumentRecord,
        negative: DocumentRecord,
        target: int,
    ):
        super().__init__(query, positive, negative)
        self.target = target


class PairwiseRecords(BaseRecords):
    """Pairwise records of queries associated with (positive, negative) pairs"""

    # The queries
    _topics: List[TopicRecord]

    # The document IDs (positive)
    positives: List[DocumentRecord]

    # The scores of the retriever
    negatives: List[DocumentRecord]

    def __init__(self):
        self._topics = []
        self.positives = []
        self.negatives = []

    def add(self, record: PairwiseRecord):
        self._topics.append(record.query)
        self.positives.append(record.positive)
        self.negatives.append(record.negative)

    @property
    def topics(self):
        return itertools.chain(self._topics, self._topics)

    def set_unique_topics(self, topics: List[TopicRecord]):
        assert len(topics) == len(
            self._topics
        ), f"Number of topics do not match ({len(topics)} vs {len(self._topics)})"
        self._topics = topics

    def set_unique_documents(self, documents: List[DocumentRecord]):
        N = len(self._topics)
        assert len(documents) == N * 2
        self.positives = documents[:N]
        self.negatives = documents[N:]

    queries = topics

    @property
    def unique_topics(self):
        return self._topics

    unique_queries = unique_topics

    @property
    def documents(self):
        return itertools.chain(self.positives, self.negatives)

    def pairs(self):
        indices = list(range(len(self._topics)))
        return indices * 2, list(range(2 * len(self.positives)))

    def __len__(self):
        return len(self._topics)

    def __getitem__(self, ix: Union[slice, int]):
        if isinstance(ix, slice):
            records = PairwiseRecords()
            for i in range(ix.start, min(ix.stop, len(self._topics)), ix.step or 1):
                records.add(
                    PairwiseRecord(
                        self._topics[i], self.positives[i], self.negatives[i]
                    )
                )
            return records

        return PairwiseRecord(self._topics[ix], self.positives[ix], self.negatives[ix])


class PairwiseRecordsWithTarget(PairwiseRecords):
    """Pairwise records associated with a label (saying which document is better)"""

    target: List[int]

    def __init__(self):
        super().__init__()
        self.target = []

    def add(self, record: PairwiseRecordWithTarget):
        self._topics.append(record.query)
        self.positives.append(record.positive)
        self.negatives.append(record.negative)
        self.target.append(record.target)

    def get_target(self):
        return self.target

    def __getitem__(self, ix: Union[slice, int]):
        if isinstance(ix, slice):
            records = PairwiseRecordsWithTarget()
            for i in range(ix.start, min(ix.stop, len(self._topics)), ix.step or 1):
                records.add(
                    PairwiseRecordWithTarget(
                        self._topics[i],
                        self.positives[i],
                        self.negatives[i],
                        self.target[i],
                    )
                )
            return records

        return PairwiseRecordWithTarget(
            self._topics[ix], self.positives[ix], self.negatives[ix], self.target[ix]
        )


class BatchwiseRecords(BaseRecords):
    """Several documents (with associated [pseudo]relevance) per query

    Assumes that the number of documents per query is always the same (even
    though documents themselves can be different)
    """

    relevances: torch.Tensor

    def __getitem__(self, ix: Union[slice, int]):
        """Sub-sample"""
        raise NotImplementedError(f"__getitem__() in {self.__class__}")


class ProductRecords(BatchwiseRecords):
    """Computes the score for all the documents and queries

    The relevance matrix

    Attributes:

        _topics: The list of queries
        _documents: The list of documents
        _relevances: (query x document) matrix with relevance score (between 0 and 1)
    """

    _topics: List[TopicRecord]
    """The list of queries to score"""

    _documents: List[DocumentRecord]
    """The list of documents to score"""

    relevances: torch.Tensor
    """A 2D tensor (query x document) indicating the relevance of the each
    query/document pair"""

    is_product = True

    def __init__(self):
        self._topics = []
        self._documents = []

    def add_topics(self, *topics: TopicRecord):
        self._topics.extend(topics)

    def add_documents(self, *documents: DocumentRecord):
        self._documents.extend(documents)

    def set_relevances(self, relevances: torch.Tensor):
        assert relevances.shape[0] == len(self._topics), (
            f"The number of queries {len(self._topics)} "
            + "does not match the number of rows {relevances.shape[0]}"
        )
        assert relevances.shape[1] == len(self._documents), (
            f"The number of documents {len(self._documents)} "
            + "does not match the number of columns {relevances.shape[1]}"
        )
        self.relevances = relevances

    def __len__(self):
        return len(self._topics)

    @property
    def topics(self):
        for q in self._topics:
            for _ in self._documents:
                yield q

    queries = topics

    @property
    def unique_topics(self):
        return self._topics

    unique_queries = unique_topics

    @property
    def documents(self):
        for _ in self._topics:
            for d in self._documents:
                yield d

    @property
    def unique_documents(self):
        return self._documents

    def pairs(self) -> Tuple[Iterable[int], Iterable[int]]:
        topics = []
        documents = []
        for q in range(len(self._topics)):
            for d in range(len(self._documents)):
                topics.append(q)
                documents.append(d)
        return topics, documents


class DocumentRecords(List[DocumentRecord]):
    """Masked Language Modeling Records are a set of documents"""

    # Text of the documents
    documents: List[DocumentRecord]

    def __init__(self):
        super().__init__()
        self.documents = []

    def add(self, record: DocumentRecord):
        self.documents.append(record)

    def __len__(self):
        return len(self.documents)

    def __getitem__(self, ix: Union[slice, int]):
        if isinstance(ix, slice):
            records = DocumentRecord()
            for i in range(ix.start, min(ix.stop, len(self)), ix.step or 1):
                records.add(self.documents[i])
            return records

        return DocumentRecords(self.documents[ix])

    @staticmethod
    def from_texts(
        documents: List[str],
    ):
        records = DocumentRecords()
        records.documents = list(documents)
        return records

    def to_texts(self) -> List[str]:
        texts = []
        for doc in self.documents:
            texts.append(doc.document[TextItem].text)

        return texts



import math
import numpy as np
from typing import Optional, List
from experimaestro import Config, Param, default
import torch
from torch import nn
from typing_extensions import Annotated
from xpmir.index import Index
import xpmir.neural.modules as modules
from xpmir.neural.interaction import (
    InteractionScorer,
    SimilarityOutput,
    TrainerContext,
    TokenizedTextEncoderBase,
    TokenizerOptions,
    TokensEncoderOutput,
)
from xpmir.text.encoders import InputType
from .common import SimilarityInputWithTokens

# The code below is heavily borrowed from OpenNIR


class CountHistogram(Config, nn.Module):
    """Base histogram class

    Attributes:
        nbins: number of bins in matching histogram
    """

    nbins: Param[int] = 29

    def forward(self, simmat: torch.Tensor, dlens: List[int], mask: torch.BoolTensor):
        """Computes the histograms for each query term

        :param simmat: A (B... x Lq x Ld) matrix
        :param mask: A (B... x Lq x Ld) mask
        :param dlens: The document lengths (vector of size D)
        :return: A (B... x Lq x nbins) matrix containing counts
        """
        # +1e-5 to nudge scores of 1 to above threshold
        bins = ((simmat + 1.00001) / 2.0 * (self.nbins - 1)).long()
        weights = mask.float()

        hist = torch.zeros(
            *simmat.shape[:-1], self.nbins, device=simmat.device, dtype=simmat.dtype
        )
        return hist.scatter_add_(simmat.ndim - 1, bins, weights)


class NormalizedHistogram(CountHistogram):
    def forward(self, simmat, dlens, mask):
        result = super().forward(simmat, dlens, mask)
        BATCH, QLEN, _ = simmat.shape
        return result / dlens.reshape(BATCH, 1).expand(BATCH, QLEN)


class LogCountHistogram(CountHistogram):
    def forward(self, simmat, dlens, mask):
        result = super().forward(simmat, dlens, mask)
        return (result.float() + 1e-5).log()


class Combination(Config, nn.Module):
    def forward(self, scores: torch.Tensor, idf: torch.Tensor):
        """Combines term scores with IDF

        :param scores: A (B... x Lq) tensor
        :param idf: A (B... x Lq) tensor
        """
        ...


class SumCombination(Combination):
    def forward(self, scores, idf):
        return scores.sum(dim=-1)


class IdfCombination(Combination):
    def forward(self, scores: torch.Tensor, idf: torch.Tensor):
        idf = idf.softmax(dim=-1)
        return (scores * idf).sum(dim=-1)


class Drmm(InteractionScorer):
    """Deep Relevance Matching Model (DRMM)

    Implementation of the DRMM model from:

      Jiafeng Guo, Yixing Fan, Qingyao Ai, and William Bruce Croft. 2016. A Deep
      Relevance Matching Model for Ad-hoc Retrieval. In CIKM.
    """

    hist: Annotated[CountHistogram, default(LogCountHistogram())]
    """The histogram type"""

    hidden: Param[int] = 5
    """Hidden layer dimension for the feed forward matching network"""

    index: Param[Optional[Index]]
    """The index (only used when using IDF to combine)"""

    combine: Annotated[Combination, default(IdfCombination())]
    """How to combine the query term scores"""

    def __validate__(self):
        super().__validate__()
        assert (self.combine != "idf") or (
            self.index is not None
        ), "index must be provided if using IDF"

    def __initialize__(self, options):
        super().__initialize__(options)
        self.simmat = modules.InteractionMatrix(self.encoder.pad_tokenid)
        self.hidden_1 = nn.Linear(self.hist.nbins, self.hidden)
        self.hidden_2 = nn.Linear(self.hidden, 1)
        self.needs_idf = isinstance(self.combine, IdfCombination)

    def _encode(
        self,
        texts: List[InputType],
        encoder: TokenizedTextEncoderBase[InputType, TokensEncoderOutput],
        options: TokenizerOptions,
    ) -> SimilarityInputWithTokens:
        encoded = encoder(texts, options=options)

        max_len = max(encoded.tokenized.lens)
        padded_tokens = [
            (t + [""] * (max_len - len(t))) for t in encoded.tokenized.tokens
        ]

        return self.similarity.preprocess(
            SimilarityInputWithTokens(
                encoded.value,
                encoded.tokenized.mask,
                np.array(padded_tokens, dtype=str),
            )
        )

    def compute_scores(
        self,
        queries: SimilarityInputWithTokens,
        documents: SimilarityInputWithTokens,
        value: SimilarityOutput,
        info: Optional[TrainerContext] = None,
    ):
        """Compute the scores given the tensor of similarities (B x Lq x Ld) or
        (Bq x Lq x Bd x Ld)"""
        # Computes the IDF if needed
        query_idf = None
        if self.needs_idf:
            assert self.index is not None
            query_idf = torch.full_like(queries.mask, float("-inf"), dtype=torch.float)
            log_nd = math.log(self.index.documentcount + 1)
            for i, tokens_i in enumerate(queries.tokens):
                for j, t in enumerate(tokens_i):
                    query_idf[i, j] = log_nd - math.log(self.index.term_df(t) + 1)

        mask = value.q_view(queries.mask) * value.d_view(documents.mask)
        similarity = value.similarity
        dlens = torch.LongTensor([len(tokens) for tokens in documents.tokens])

        if value.similarity.ndim == 4:
            # Transform into B... x Lq x Ld shape
            similarity = value.similarity.transpose(1, 2)
            mask = mask.transpose(1, 2)
            query_idf = query_idf.unsqueeze(0)

        qterm_features = self.histogram_pool(similarity, dlens, mask)
        qterm_scores = self.hidden_2(torch.relu(self.hidden_1(qterm_features))).squeeze(
            -1
        )
        return self.combine(qterm_scores, query_idf)

    def histogram_pool(self, simmat, dlens, mask):
        histogram = self.hist(simmat, dlens, mask)
        return histogram



import torch


class BCEWithLogLoss(torch.autograd.Function):
    """Binary cross-entropy loss when outputs are log probabilities

    Both arguments should be vectors

    First argument: log_probs must be strictly negative
    Second arguments: targets, can be betwen 0 and 1
    """

    @staticmethod
    def forward(ctx, log_probs: torch.Tensor, targets: torch.Tensor):
        assert torch.all(log_probs < 0.0)

        # Computes the loss
        p_neg = 1.0 - log_probs.exp()
        loss = (
            -(targets * log_probs).sum()
            - ((1 - targets) * (1.0 - log_probs.exp()).log()).sum()
        )

        # Save for backward
        ctx.save_for_backward(log_probs, p_neg, targets)

        return loss / log_probs.numel()

    @staticmethod
    def backward(ctx, grad_output):
        if not ctx.needs_input_grad[0]:
            return None

        log_probs, p_neg, targets = ctx.saved_tensors

        # Stabilizes the gradient (see Loss.cpp in Aten)
        p_neg = p_neg.clamp_min(1e-12)
        grad_output = grad_output / log_probs.numel()

        grad_log_probs = -targets + (targets - 1) * (1.0 - 1.0 / p_neg)
        return grad_output * grad_log_probs, None


bce_with_logits_loss = BCEWithLogLoss.apply



from typing import (
    Iterable,
    Iterator,
    NamedTuple,
    Tuple,
    List,
    Any,
)

import numpy as np
from datamaestro.data import File
from datamaestro_text.data.ir.base import (
    TopicRecord,
    DocumentRecord,
    ScoredItem,
    SimpleTextItem,
    IDItem,
    create_record,
)
from experimaestro import Config, Meta, Param
from xpmir.learning import Sampler
from xpmir.letor.samplers.hydrators import SampleHydrator
from xpmir.rankers import ScoredDocument
from xpmir.utils.iter import (
    SerializableIterator,
    SkippingIterator,
    SerializableIteratorTransform,
)


class PairwiseDistillationSample(NamedTuple):
    query: TopicRecord
    """The query"""

    documents: Tuple[DocumentRecord, DocumentRecord]
    """Positive/negative document with teacher scores"""


class PairwiseDistillationSamples(Config, Iterable[PairwiseDistillationSample]):
    """Pairwise distillation file"""

    def __iter__(self) -> Iterator[PairwiseDistillationSample]:
        raise NotImplementedError()


class PairwiseHydrator(PairwiseDistillationSamples, SampleHydrator):
    """Hydrate ID-based samples with document and/or query content"""

    samples: Param[PairwiseDistillationSamples]
    """The distillation samples without texts for query and documents"""

    def transform(self, sample: PairwiseDistillationSample):
        topic, documents = sample.query, sample.documents

        if transformed := self.transform_topics(topic):
            topic = transformed[0]

        if transformed := self.transform_documents(documents):
            documents = tuple(
                ScoredDocument(d, sd[ScoredItem].score)
                for d, sd in zip(transformed, sample.documents)
            )

        return PairwiseDistillationSample(topic, documents)

    def __iter__(self) -> Iterator[PairwiseDistillationSample]:
        iterator = iter(self.samples)
        return SerializableIteratorTransform(
            SkippingIterator.make_serializable(iterator), self.transform
        )


class PairwiseDistillationSamplesTSV(PairwiseDistillationSamples, File):
    """A TSV file (Score 1, Score 2, Query, Document 1, Document 2)"""

    with_docid: Meta[bool]
    with_queryid: Meta[bool]

    def __iter__(self) -> Iterator[PairwiseDistillationSample]:
        return self.iter()

    def iter(self) -> Iterator[PairwiseDistillationSample]:
        import csv

        def iterate():
            with self.path.open("rt") as fp:
                for row in csv.reader(fp, delimiter="\t"):
                    if self.with_queryid:
                        query = create_record(id=row[2])
                    else:
                        query = create_record(text=row[2])

                    if self.with_docid:
                        documents = (
                            DocumentRecord(IDItem(row[3]), ScoredItem(float(row[0]))),
                            DocumentRecord(IDItem(row[4]), ScoredItem(float(row[1]))),
                        )
                    else:
                        documents = (
                            DocumentRecord(
                                SimpleTextItem(row[3]), ScoredItem(float(row[0]))
                            ),
                            DocumentRecord(
                                SimpleTextItem(row[4]), ScoredItem(float(row[1]))
                            ),
                        )

                    yield PairwiseDistillationSample(query, documents)

        return SkippingIterator(iterate())


class DistillationPairwiseSampler(Sampler):
    """Just loops over samples"""

    samples: Param[PairwiseDistillationSamples]

    def initialize(self, random: np.random.RandomState):
        super().initialize(random)

    def pairwise_iter(self) -> SerializableIterator[PairwiseDistillationSample, Any]:
        return SkippingIterator.make_serializable(iter(self.samples))

    def pairwise_batch_iter(
        self, size
    ) -> SerializableIterator[List[PairwiseDistillationSample], Any]:
        """Batchwise iterator

        Can be subclassed by some classes to be more efficient"""

        class BatchIterator(SerializableIterator):
            def __init__(self, sampler: DistillationPairwiseSampler):
                self.iter = sampler.pairwise_iter()

            def state_dict(self):
                return self.iter.state_dict()

            def load_state_dict(self, state):
                self.iter.load_state_dict(state)

            def __next__(self):
                batch = []
                for _, record in zip(range(size), self.iter):
                    batch.append(record)
                return batch

        return BatchIterator(self)



import numpy as np
from typing import Callable
import logging
from decorator import decorator
from xpmir.utils.iter import (
    RandomSerializableIterator,
    SkippingIterator,
    InfiniteSkippingIterator,
    MultiprocessSerializableIterator,
    RandomizedSerializableIterator,
    SerializableIterator,
    RandomStateSerializableAdaptor,
)


@decorator
def iter_checker(
    factory: Callable[..., SerializableIterator], steps: int = 5, *args, **kw
):
    # Iterate and get state
    iterator = factory(*args, **kw)
    for _ in range(steps):
        next(iterator)

    state = iterator.state_dict()

    # Get the next value
    x = next(iterator)

    # Re-create and test
    iterator = factory()
    iterator.load_state_dict(state)
    assert next(iterator) == x


@iter_checker(steps=5)
def test_iter_skipping_iterator():
    return SkippingIterator(iter(list(range(10))))


@iter_checker(steps=5)
def test_iter_skipping_infinite_iterator():
    return InfiniteSkippingIterator(list(range(3)))


@iter_checker(steps=5)
def test_iter_mp_iterator():
    """Test the multiprocess iterator"""
    return MultiprocessSerializableIterator(SkippingIterator(iter(list(range(10)))))


@iter_checker(steps=5)
def test_iter_randomized_iterator():
    class Iterator(RandomStateSerializableAdaptor[InfiniteSkippingIterator]):
        def __init__(self, n: int):
            base = InfiniteSkippingIterator(list(range(n)))
            super().__init__(base)

        def __next__(self):
            element = next(self.iterator)
            randint = self.random.randint(10)
            logging.debug(
                "---> [%s] %s + %s",
                hash(str(self.random.get_state())),
                element,
                randint,
            )
            return element + randint

    rs = np.random.RandomState()
    return RandomizedSerializableIterator(rs, Iterator(3))


@iter_checker(steps=5)
def test_iter_random_iterator():
    a = list(range(10))

    def create_iter(state: np.random.RandomState):
        while True:
            ix = state.randint(len(a))
            yield a[ix]

    rs = np.random.RandomState()
    return RandomSerializableIterator(rs, create_iter)



import torch
from torch.autograd import gradcheck
from xpmir.learning.losses import bce_with_logits_loss


def test_pairwise_bce_loss():
    input = -torch.randn(
        (12,), dtype=torch.double, requires_grad=True
    ).abs(), torch.randint(0, 2, (12,))
    gradcheck(bce_with_logits_loss, input, eps=1e-6, atol=1e-4)



from xpmir.neural.modules.interaction_matrix import (
    InteractionMatrix,
    cos_simmat,
    binmat,
)

# from xpmir.neural.modules.bert import CustomBertModel, CustomBertModelWrapper
# from xpmir.neural.modules.rbf_kernels import RbfKernelBank



from abc import abstractmethod
import itertools
from typing import Iterable, Union, List, Optional, TypeVar, Generic, Sequence
import torch
from datamaestro_text.data.ir import TextItem
from xpmir.learning.context import TrainerContext
from xpmir.letor.records import BaseRecords, ProductRecords, TopicRecord, DocumentRecord
from xpmir.rankers import LearnableScorer

QueriesRep = TypeVar("QueriesRep", bound=Sequence)
DocsRep = TypeVar("DocsRep", bound=Sequence)


class DualRepresentationScorer(LearnableScorer, Generic[QueriesRep, DocsRep]):
    """Neural scorer based on (at least a partially) independent representation
    of the document and the question.

    This is the base class for all scorers that depend on a map
    of cosine/inner products between query and document tokens.
    """

    def forward(self, inputs: BaseRecords, info: Optional[TrainerContext] = None):
        # Forward to model
        enc_queries = self.encode_queries(list(inputs.unique_queries))
        enc_documents = self.encode_documents(list(inputs.unique_documents))

        # Score product
        if isinstance(inputs, ProductRecords):
            return self.score_product(
                enc_queries,
                enc_documents,
                info,
            ).flatten()

        # Score pairs
        pairs = inputs.pairs()
        q_ix, d_ix = pairs
        return self.score_pairs(
            enc_queries[
                q_ix
            ],
            enc_documents[
                d_ix
            ],
            info,
        ).flatten()

    def encode(self, texts: Iterable[str]) -> Union[DocsRep, QueriesRep]:
        """Encode a list of texts (document or query)

        The return value is model dependent"""
        raise NotImplementedError()

    def encode_documents(self, records: Iterable[DocumentRecord]) -> DocsRep:
        """Encode a list of texts (document or query)

        The return value is model dependent"""
        return self.encode([record[TextItem].text for record in records])

    def encode_queries(self, records: Iterable[TopicRecord]) -> QueriesRep:
        """Encode a list of texts (document or query)

        The return value is model dependent, but should be sequence

        By default, uses `merge`
        """
        return self.encode([record[TextItem].text for record in records])

    def merge_queries(self, queries: QueriesRep):
        """Merge query batches encoded with `encode_queries`

        By default, uses `merge`
        """
        return self.merge(queries)

    def merge_documents(self, documents: DocsRep):
        """Merge query batches encoded with `encode_documents`"""
        return self.merge(documents)

    def merge(self, objects: Union[DocsRep, QueriesRep]):
        """Merge objects

        - for tensors, uses torch.cat
        - for lists, concatenate all of them
        """
        assert isinstance(
            objects, List
        ), f"Merging can only be done with lists, got {type(objects)}"

        # Just returns the only object to merge
        if len(objects) == 1:
            return objects[0]

        if isinstance(objects[0], torch.Tensor):
            return torch.cat(objects)

        if isinstance(objects[0], List):
            return list(itertools.chain(objects))

        from xpmir.text.encoders import TextsRepresentationOutput
        from xpmir.text.tokenizers import TokenizedTexts

        if isinstance(objects[0], TextsRepresentationOutput):

            def merge_mask(mask):
                min_batch_size = torch.min(
                    torch.tensor(list(map(lambda x: x.shape, mask))), dim=0
                )[0][0]
                batch_equalized = list(
                    itertools.chain(
                        *map(lambda x: x.t().split(min_batch_size, dim=1), mask)
                    )
                )

                pad = torch.nn.utils.rnn.pad_sequence(batch_equalized)

                return pad.reshape(pad.shape[0], -1).t()

            tokenized = list(map(lambda x: x.tokenized, objects))

            tokens = list(
                filter(lambda x: x is not None, map(lambda x: x.tokens, tokenized))
            )
            tokens = None if len(tokens) == 0 else list(itertools.chain(*tokens))

            ids = merge_mask(list(map(lambda x: x.ids, tokenized)))

            lens = list(itertools.chain(*map(lambda x: x.lens, tokenized)))

            mask = list(map(lambda x: x.mask, tokenized))
            mask = None if len(mask) == 0 else merge_mask(mask)

            token_type_ids = list(
                filter(
                    lambda x: x is not None, map(lambda x: x.token_type_ids, tokenized)
                )
            )
            token_type_ids = (
                None if len(token_type_ids) == 0 else torch.cat(token_type_ids)
            )

            return TextsRepresentationOutput(
                torch.cat(list(map(lambda x: x.value, objects))),
                TokenizedTexts(tokens, ids, lens, mask, token_type_ids),
            )
        raise RuntimeError(f"Cannot deal with objects of type {type(list[0])}")

    @abstractmethod
    def score_product(
        self,
        queries: QueriesRep,
        documents: DocsRep,
        info: Optional[TrainerContext] = None,
    ) -> torch.Tensor:
        """Computes the score of all possible pairs of query and document

        Args:
            queries (Any): The encoded queries
            documents (Any): The encoded documents
            info (Optional[TrainerContext]): The training context (if learning)

        Returns:
            torch.Tensor:
                A tensor of dimension (N, P) where N is the number of queries
                and P the number of documents
        """
        ...

    @abstractmethod
    def score_pairs(
        self,
        queries: QueriesRep,
        documents: DocsRep,
        info: Optional[TrainerContext] = None,
    ) -> torch.Tensor:
        """Score the specified pairs of queries/documents.

        There are as many queries as documents. The exact type of
        queries and documents depends on the specific instance of the
        dual representation scorer.

        Args:
            queries (QueriesRep): The list of encoded queries
            documents (DocsRep): The matching list of encoded documents
            info (Optional[TrainerContext]): _description_

        Returns:
            torch.Tensor:
                A tensor of dimension (N) where N is the number of documents/queries
        """
        ...



import logging
from experimaestro import Param
from xpmir.learning.context import TrainState, InitializationTrainingHook
from xpmir.learning.parameters import ParametersIterator
from xpmir.utils.utils import easylog

logger = easylog()
logger.setLevel(logging.INFO)


class LayerFreezer(InitializationTrainingHook):
    """This training hook class can be used to freeze a subset of model
    parameters"""

    selector: Param[ParametersIterator]
    """How to select the layers to freeze"""

    def __init__(self):
        self._initialized = False

    def after(self, state: TrainState):
        if not self._initialized:
            self._initialized = True
            for name, module, param, to_freeze in self.selector.iter():
                if to_freeze:
                    logger.info("Freezing layer %s", name)
                    param.requires_grad = False


class LayerSharer(InitializationTrainingHook):
    """This training hook class can be used to share parameters"""

    source: Param[ParametersIterator]
    """The parameters to share"""

    target: Param[ParametersIterator]
    """The parameters to be shared"""

    def __init__(self):
        self._initialized = False

    def after(self, state: TrainState):
        if not self._initialized:
            self._initialized = True
            for source, target in zip(
                self.source.selected(), self.target.selected(), strict=True
            ):
                logger.info("Sharing layer %s -> %s", source.name, target.name)
                target.set(source.parameter)



from .encoders import (  # noqa: F401
    TokensEncoder,
    TextEncoderBase,
    TokensEncoderOutput,
    TokenizedTextEncoderBase,
    TokenizedTextEncoder,
    InputType,
)

from .tokenizers import (  # noqa: F401
    TokenizerBase,
    TokenizerOptions,
    TokenizerOutput,
    TokenizerInput,
    TokenizedTexts,
)



"""
Pre-computed Anserini indices provided by Jimmy Lin (U. Waterloo)
"""

from datamaestro.definitions import argument, datatasks, datatags, dataset
from datamaestro.download.archive import tardownloader
from datamaestro.annotations.agreement import useragreement
from datamaestro.utils import HashCheck
from xpmir.index.anserini import Index
from hashlib import md5


@useragreement(
    """Robust04 dataset.
Please confirm you agree to the authors' data usage stipulations found at
https://trec.nist.gov/data/cd45/index.html"""
)
@tardownloader(
    "index",
    url="https://git.uwaterloo.ca/jimmylin/anserini-indexes/raw/master/index-robust04-20191213.tar.gz",
    checker=HashCheck("15f3d001489c97849a010b0a4734d018", md5),
)
@dataset(Index)
def robust04(index):
    """Robust 2014 index

    Pre-computed Anserini index of the Robust 2014 collection; used parameters can be found
    at https://git.uwaterloo.ca/jimmylin/anserini-indexes/-/blob/master/index-robust04-20191213-readme.txt
    """
    return {
        "path": index,
        "storePositions": True,
        "storeDocvectors": True,
        "storeRaw": True,
        "storeContents": False,
        "stemmer": "porter",
    }



"""Returns the TAS-Balanced model (from huggingface)

> Efficiently Teaching an Effective Dense Retriever with Balanced Topic Aware Sampling
> Sebastian Hofstätter, Sheng-Chieh Lin, Jheng-Hong Yang, Jimmy Lin, Allan Hanbury

Returns:
    DotDense: A DotDense ranker based on tas-balanced
"""
from xpmir.text.huggingface import TransformerEncoder, TransformerTextEncoderAdapter
from experimaestro.huggingface import ExperimaestroHFHub
from xpmir.neural.dual import DotDense

encoder = TransformerEncoder(
    model_id="sebastian-hofstaetter/distilbert-dot-tas_b-b256-msmarco",
    trainable=True,
    maxlen=200,
)

model = DotDense(
    encoder=encoder,
    query_encoder=TransformerTextEncoderAdapter(encoder=encoder, maxlen=30),
)

ExperimaestroHFHub(model).push_to_hub(repo_id="xpmir/tas-balanced")



from experimaestro import Param, Constant
from datamaestro_text.data.ir import TextItem
from xpmir.neural.generative import ConditionalGenerator
from xpmir.letor.records import (
    BaseRecords,
)
from xpmir.learning.context import TrainerContext
from xpmir.rankers import LearnableScorer, ScorerOutputType


class GenerativeCrossScorer(LearnableScorer):
    """A cross-encoder based on a generative model"""

    version: Constant[int] = 2
    """Generative cross scorer version

    changelog:

        1. corrects output type probability
    """

    outputType: ScorerOutputType = ScorerOutputType.LOG_PROBABILITY

    #: The pattern used to condition the decoder, with query / document replaced
    # by their string values
    pattern: Param[str] = "Query: {query} Document: {document} Relevant:"

    #: Conditional generator
    generator: Param[ConditionalGenerator]

    #: Relevant token ID
    relevant_token_id: Param[int]

    def __initialize__(self, options):
        super().__initialize__(options)
        self.generator.initialize(options)

    def forward(self, inputs: BaseRecords, info: TrainerContext = None):
        # Encode queries and documents
        inputs = [
            self.pattern.format(query=tr[TextItem].text, document=dr[TextItem].text)
            for tr, dr in zip(inputs.topics, inputs.documents)
        ]

        step_generator = self.generator.stepwise_iterator()
        step_generator.init(inputs)
        logits = step_generator.step(None)
        return logits.log_softmax(dim=1)[:, self.relevant_token_id]



import sys
import types
from xpmir.utils.utils import easylog

_logger = easylog()


# hack for pylint
def FP16_Optimizer(*args, **kwargs):
    return None


def FusedAdam(*args, **kwargs):
    return None


class ApexWrapper(types.ModuleType):
    @property
    def _apex(self):
        try:
            import apex

            return apex
        except ImportError:
            _logger.warning(
                "Module apex not installed. Please see <https://github.com/NVIDIA/apex>"
            )
            raise

    @property
    def FP16_Optimizer(self):
        return self._apex.optimizers.FP16_Optimizer

    @property
    def FusedAdam(self):
        return self._apex.optimizers.FusedAdam


sys.modules[__name__].__class__ = ApexWrapper



from experimaestro.launcherfinder import find_launcher

from xpmir.papers import configuration, attrs_cached_property


@configuration()
class LauncherSpecification:
    """Launcher specification

    This allows requesting computational resources such as 2 GPUs with more than
    12Go of memory)
    """

    requirements: str

    @attrs_cached_property
    def launcher(self):
        return find_launcher(self.requirements)



from typing import List
from functools import cached_property
from xpmir.learning.optim import (
    AdamW,
    Adam,
    Adafactor,
    SGD,
    ParameterOptimizer,
    RegexParameterFilter,
    get_optimizers,
)
from xpmir.learning.schedulers import LinearWithWarmup
from xpmir.papers import configuration


@configuration
class TransformerOptimization:
    """Configuration for a transformer optimization"""

    scheduler: bool = True
    warmup_min_factor: float = 0

    num_warmup_steps: int = 1000
    batch_size: int = 64
    max_epochs: int = 3200
    steps_per_epoch: int = 32
    """Number of steps (batches) per epoch"""

    optimizer_name: str = "adam-w"
    lr: float = 3.0e-6
    weight_decay: float = 1e-2
    eps: float = 1e-8

    re_no_l2_regularization: List[str] = [r"\.bias$", r"\.LayerNorm\."]
    """Regular expression for layers (targets BERT parameters)"""

    def get_optimizer(self, regularization):
        # Set weight decay to 0 if no regularization
        weight_decay = self.weight_decay if regularization else 0

        if self.optimizer_name == "adam-w":
            return AdamW(
                lr=self.lr,
                weight_decay=weight_decay,
                eps=self.eps,
            )
        elif self.optimizer_name == "adam":
            return Adam(self.lr, weight_decay=weight_decay, eps=self.eps)
        elif self.optimizer_name == "sgd":
            return SGD(lr=self.lr, weight_decay=weight_decay)
        elif self.optimizer_name == "adafactor":
            return Adafactor(
                lr=self.lr, weight_decay=weight_decay, relative_step=self.lr is None
            )
        else:
            raise ValueError(f"Cannot handle optimizer named {self.optimizer_Name}")

    @cached_property
    def optimizer(self):
        scheduler = (
            LinearWithWarmup(
                num_warmup_steps=self.num_warmup_steps,
                min_factor=self.warmup_min_factor,
            )
            if self.scheduler
            else None
        )

        if not self.re_no_l2_regularization:
            return get_optimizers(
                [
                    ParameterOptimizer(
                        scheduler=scheduler,
                        optimizer=self.get_optimizer(True),
                    ),
                ]
            )

        return get_optimizers(
            [
                ParameterOptimizer(
                    scheduler=scheduler,
                    optimizer=self.get_optimizer(False),
                    filter=RegexParameterFilter(includes=self.re_no_l2_regularization),
                ),
                ParameterOptimizer(
                    scheduler=scheduler,
                    optimizer=self.get_optimizer(True),
                ),
            ]
        )



import io
import json
from pathlib import Path
from typing import Iterator, List, Tuple, Dict, Any
import numpy as np
from datamaestro.record import Record
from datamaestro_text.data.ir import (
    Adhoc,
    TrainingTriplets,
    PairwiseSampleDataset,
    PairwiseSample,
    ScoredItem,
    DocumentStore,
    TextItem,
    SimpleTextItem,
    create_record,
    DocumentRecord,
    IDItem,
)
from experimaestro import Param, tqdm, Task, Annotated, pathgenerator
from experimaestro.annotations import cache
from experimaestro.compat import cached_property
import torch
from xpmir.rankers import ScoredDocument
from xpmir.datasets.adapters import TextStore
from xpmir.letor.records import (
    BatchwiseRecords,
    PairwiseRecords,
    ProductRecords,
    PairwiseRecord,
    PointwiseRecord,
    TopicRecord,
)
from xpmir.rankers import Retriever, Scorer
from xpmir.learning import Sampler
from xpmir.utils.utils import easylog
from xpmir.utils.iter import (
    RandomSerializableIterator,
    SerializableIterator,
    SerializableIteratorAdapter,
    SkippingIterator,
    RandomStateSerializableAdaptor,
    InfiniteSkippingIterator,
    iterable_of,
)
from datamaestro_text.interfaces.plaintext import read_tsv

logger = easylog()


# --- Base classes for samplers


class PointwiseSampler(Sampler):
    def pointwise_iter(self) -> SerializableIterator[PointwiseRecord, Any]:
        """Iterable over pointwise records"""
        raise NotImplementedError(f"{self.__class__} should implement PointwiseRecord")


class PairwiseSampler(Sampler):
    """Abstract class for pairwise samplers which output a set of (query,
    positive, negative) triples"""

    def pairwise_iter(self) -> SerializableIterator[PairwiseRecord, Any]:
        """Iterate over batches of size (# of queries) batch_size

        Args:
            batch_size: Number of queries per batch
        """
        raise NotImplementedError(f"{self.__class__} should implement __iter__")

    def pairwise_batch_iter(self, size) -> SerializableIterator[PairwiseRecords, Any]:
        """Batchwise iterator

        Can be subclassed by some classes to be more efficient"""

        class BatchIterator(SerializableIterator):
            def __init__(self, sampler: PairwiseSampler):
                self.iter = sampler.pairwise_iter()

            def state_dict(self):
                return self.iter.state_dict()

            def load_state_dict(self, state):
                self.iter.load_state_dict(state)

            def __next__(self):
                batch = PairwiseRecords()
                for _, record in zip(range(size), self.iter):
                    batch.add(record)
                return batch

        return BatchIterator(self)


class BatchwiseSampler(Sampler):
    """Base class for batchwise samplers, that provide for each question a list
    of documents"""

    def batchwise_iter(
        self, batch_size: int
    ) -> SerializableIterator[BatchwiseRecords, Any]:
        """Iterate over batches of size (# of queries) batch_size

        Args:
            batch_size: Number of queries per batch
        """
        raise NotImplementedError(f"{self.__class__} should implement __iter__")


# --- Real instances


class ModelBasedSampler(Sampler):
    """Base class for retriever-based sampler"""

    dataset: Param[Adhoc]
    """The IR adhoc dataset"""

    retriever: Param[Retriever]
    """A retriever to sample negative documents"""

    _store: DocumentStore

    def __validate__(self) -> None:
        super().__validate__()

        assert self.retriever.get_store() is not None or isinstance(
            self.dataset.documents, DocumentStore
        ), "The retriever has no associated document store (to get document text)"

    def initialize(self, random):
        super().initialize(random)
        self._store = self.retriever.get_store() or self.dataset.documents
        assert self._store is not None, "No document store found"

    def document(self, doc_id):
        """Returns the document textual content"""
        return self._store.document_ext(doc_id)

    def document_text(self, doc_id):
        return self.document(doc_id).text

    @cache("run")
    def _itertopics(
        self, runpath: Path
    ) -> Iterator[
        Tuple[str, List[Tuple[str, int, float]], List[Tuple[str, int, float]]]
    ]:
        """Iterates over topics, returning retrieved positives and negatives
        documents"""
        self.logger.info("Reading topics and retrieving documents")

        if not runpath.is_file():
            tmprunpath = runpath.with_suffix(".tmp")

            with tmprunpath.open("wt") as fp:

                # Read the assessments
                self.logger.info("Reading assessments")
                assessments: Dict[str, Dict[str, float]] = {}
                for qrels in self.dataset.assessments.iter():
                    doc2rel = {}
                    assessments[qrels.topic_id] = doc2rel
                    for qrel in qrels.assessments:
                        doc2rel[qrel.doc_id] = qrel.rel
                self.logger.info("Read assessments for %d topics", len(assessments))

                self.logger.info("Retrieving documents for each topic")
                queries = []
                for query in self.dataset.topics.iter():
                    queries.append(query)

                # Retrieve documents
                skipped = 0
                for query in tqdm(queries):
                    q_fp = io.StringIO()
                    qassessments = assessments.get(query[IDItem].id, None)
                    if not qassessments:
                        skipped += 1
                        self.logger.warning(
                            "Skipping topic %s (no assessments)", query[IDItem].id
                        )
                        continue

                    # Write all the positive documents
                    positives = []
                    for docno, rel in qassessments.items():
                        if rel > 0:
                            q_fp.write(
                                f"{query.text if not positives else ''}"
                                f"\t{docno}\t0.\t{rel}\n"
                            )
                            positives.append((docno, rel, 0))

                    if not positives:
                        self.logger.warning(
                            "Skipping topic %s (no relevant documents)",
                            query[IDItem].id,
                        )
                        skipped += 1
                        continue

                    scoreddocuments: List[ScoredDocument] = self.retriever.retrieve(
                        query.text
                    )

                    negatives = []
                    for rank, sd in enumerate(scoreddocuments):
                        # Get the assessment (assumes not relevant)
                        rel = qassessments.get(sd.document[IDItem].id, 0)
                        if rel > 0:
                            continue

                        negatives.append((sd.document[IDItem].id, rel, sd.score))
                        q_fp.write(f"\t{sd.document[IDItem].id}\t{sd.score}\t{rel}\n")

                    if not negatives:
                        self.logger.warning(
                            "Skipping topic %s (no negatives documents)",
                            query[IDItem].id,
                        )
                        skipped += 1
                        continue

                    assert len(positives) > 0 and len(negatives) > 0

                    # Write in cache, and yield
                    fp.write(q_fp.getvalue())
                    q_fp.close()
                    yield query.text, positives, negatives

                # Finally, move the cache file in place...
                self.logger.info(
                    "Processed %d topics (%d skipped)", len(queries), skipped
                )
                tmprunpath.rename(runpath)
        else:
            # Read from cache
            self.logger.info("Reading records from file %s", runpath)
            with runpath.open("rt") as fp:
                positives = []
                negatives = []
                oldtitle = ""

                for line in fp.readlines():
                    title, docno, score, rel = line.rstrip().split("\t")
                    if title:
                        if oldtitle:
                            yield oldtitle, positives, negatives
                        positives = []
                        negatives = []
                    else:
                        title = oldtitle
                    title = title or oldtitle
                    rel = int(rel)
                    (positives if rel > 0 else negatives).append(
                        (docno, rel, float(score))
                    )
                    oldtitle = title

                yield oldtitle, positives, negatives


class PointwiseModelBasedSampler(PointwiseSampler, ModelBasedSampler):
    relevant_ratio: Param[float] = 0.5
    """The target relevance ratio"""

    def initialize(self, random):
        super().initialize(random)

        self.retriever.initialize()
        self.pos_records, self.neg_records = self.readrecords()
        self.logger.info(
            "Loaded %d/%d pos/neg records", len(self.pos_records), len(self.neg_records)
        )

    def prepare(self, sample: Tuple[str, int, float]):
        assert self.document_text(sample[1]) is not None
        document = self.document_text(sample[1])

        return PointwiseRecord(
            topic=TopicRecord(SimpleTextItem(sample[0])),
            document=DocumentRecord(document=document),
            relevance=sample[3],
        )

    def readrecords(self, runpath=None):
        pos_records, neg_records = [], []
        for title, positives, negatives in self._itertopics():
            for docno, rel, score in positives:
                pos_records.append((title, docno, score, rel))
            for docno, rel, score in negatives:
                neg_records.append((title, docno, score, rel))

        return pos_records, neg_records

    def record_iter(self) -> Iterator[PointwiseRecord]:
        npos = len(self.pos_records)
        nneg = len(self.neg_records)
        while True:
            if self.random.random() < self.relevant_ratio:
                yield self.prepare(self.pos_records[self.random.randint(0, npos)])
            else:
                yield self.prepare(self.neg_records[self.random.randint(0, nneg)])

    def pointwise_iter(self) -> SerializableIterator[PointwiseRecord, Any]:
        npos = len(self.pos_records)
        nneg = len(self.neg_records)

        def iter(random):
            while True:
                if self.random.random() < self.relevant_ratio:
                    yield self.prepare(self.pos_records[self.random.randint(0, npos)])
                else:
                    yield self.prepare(self.neg_records[self.random.randint(0, nneg)])

        return RandomSerializableIterator(self.random, iter)


class PairwiseModelBasedSampler(PairwiseSampler, ModelBasedSampler):
    """A pairwise sampler based on a retrieval model"""

    def initialize(self, random: np.random.RandomState):
        super().initialize(random)

        self.retriever.initialize()
        self.topics: List[Tuple[str, List, List]] = self._readrecords()

    def _readrecords(self):
        topics = []
        for title, positives, negatives in self._itertopics():
            topics.append((title, positives, negatives))
        return topics

    def sample(self, samples: List[Tuple[str, int, float]]):
        text = None
        while text is None:
            docid, rel, score = samples[self.random.randint(0, len(samples))]
            document = self.document(docid).update(ScoredItem(score))
            text = document[TextItem].text
        return document

    def pairwise_iter(self) -> SerializableIterator[PairwiseRecord, Any]:
        def iter(random):
            while True:
                title, positives, negatives = self.topics[
                    random.randint(0, len(self.topics))
                ]
                yield PairwiseRecord(
                    create_record(text=title),
                    self.sample(positives),
                    self.sample(negatives),
                )

        return RandomSerializableIterator(self.random, iter)


class PairwiseInBatchNegativesSampler(BatchwiseSampler):
    """An in-batch negative sampler constructured from a pairwise one"""

    sampler: Param[PairwiseSampler]
    """The base pairwise sampler"""

    def initialize(self, random):
        super().initialize(random)
        self.sampler.initialize(random)

    def batchwise_iter(
        self, batch_size: int
    ) -> SerializableIterator[BatchwiseRecords, Any]:
        def iter(pair_iter):
            # Pre-compute relevance matrix (query x document)
            relevances = torch.cat(
                (torch.eye(batch_size), torch.zeros(batch_size, batch_size)), 1
            )

            while True:
                batch = ProductRecords()
                positives = []
                negatives = []
                for _, record in zip(range(batch_size), pair_iter):
                    batch.add_topics(record.query)
                    positives.append(record.positive)
                    negatives.append(record.negative)
                batch.add_documents(*positives)
                batch.add_documents(*negatives)
                batch.set_relevances(relevances)
                yield batch

        return SerializableIteratorAdapter(self.sampler.pairwise_iter(), iter)


class TripletBasedSampler(PairwiseSampler):
    """Sampler based on a triplet source"""

    source: Param[TrainingTriplets]
    """Triplets"""

    def pairwise_iter(self) -> SerializableIterator[PairwiseRecord, Any]:
        iterator = (
            PairwiseRecord(topic, pos, neg) for topic, pos, neg in self.source.iter()
        )

        return SkippingIterator(iterator)


class PairwiseDatasetTripletBasedSampler(PairwiseSampler):
    """Sampler based on a dataset where each query is associated
    with (1) a set of relevant documents (2) negative documents,
    where each negative is sampled with a specific algorithm
    """

    documents: Param[DocumentStore]
    """The document store"""

    dataset: Param[PairwiseSampleDataset]
    """The dataset which contains the generated queries with its positives and
    negatives"""

    negative_algo: Param[str] = "random"
    """The algo to sample the negatives, default value is random"""

    def pairwise_iter(self) -> SkippingIterator[PairwiseRecord]:
        class _Iterator(
            RandomStateSerializableAdaptor[SerializableIterator[PairwiseSample]]
        ):
            def __init__(
                self,
                iterator: SerializableIterator[PairwiseSample],
                random: np.random.RandomState,
                negative_algo: str,
                documents: DocumentStore,
            ):
                super().__init__(iterator)
                self.random = random
                self.negative_algo = negative_algo
                self.documents = documents

            def __next__(self):
                sample = next(self.iterator)  # type: PairwiseSample
                possible_algos = sample.negatives.keys()

                assert (
                    self.negative_algo in possible_algos
                    or self.negative_algo == "random"
                )

                pos = sample.positives[self.random.randint(len(sample.positives))]
                qry = sample.topics[self.random.randint(len(sample.topics))]

                if self.negative_algo == "random":
                    # choose the random negatives
                    while True:
                        neg_id = self.documents.docid_internal2external(
                            self.random.randint(0, self.documents.documentcount)
                        )
                        if neg_id != pos.id:
                            break
                    neg = create_record(id=neg_id)
                else:
                    negatives = sample.negatives[self.negative_algo]
                    neg = negatives[self.random.randint(len(negatives))]

                return PairwiseRecord(
                    qry.as_record(), DocumentRecord(pos), DocumentRecord(neg)
                )

        base = InfiniteSkippingIterator(iterable_of(lambda: self.dataset.iter()))

        return _Iterator(base, self.random, self.negative_algo, self.documents)


# --- Dataloader


class TSVPairwiseSampleDataset(PairwiseSampleDataset):
    """Read the pairwise sample dataset from a tsv file"""

    hard_negative_samples_path: Param[Path]
    """The path which stores the existing ids"""

    def iter(self) -> Iterator[PairwiseSample]:
        """return a iterator over a set of pairwise_samples"""
        for triplet in read_tsv(self.hard_negative_samples_path):
            topics = [triplet[0]]
            positives = triplet[2].split(" ")
            negatives = triplet[4].split(" ")
            # at the moment, I don't have some good idea to store the algo
            yield PairwiseSample(topics=topics, positives=positives, negatives=negatives)


class JSONLPairwiseSampleDataset(PairwiseSampleDataset):
    """Transform a jsonl file to a pairwise dataset
    General format:
    {
        queries: [str, str],
        pos_ids: [id, id],
        neg_ids: {
            "bm25": [id, id],
            "random": [id, id]
        }
    }
    """

    path: Param[Path]
    """The path to the Jsonl file"""

    @cached_property
    def count(self):
        with self.path.open("r") as fp:
            line_count = sum(1 for _ in fp)
        return line_count

    def iter(self) -> Iterator[PairwiseSample]:
        with self.path.open("r") as fp:
            for line in fp:
                sample = json.loads(line)
                topics = []
                positives = []
                negatives = {}
                for topic_text in sample["queries"]:
                    topics.append(create_record(text=topic_text))
                for pos_id in sample["pos_ids"]:
                    positives.append(create_record(id=pos_id))
                for algo in sample["neg_ids"].keys():
                    negatives[algo] = []
                    for neg_id in sample["neg_ids"][algo]:
                        negatives[algo].append(create_record(id=neg_id))
                yield PairwiseSample(
                    topics=topics, positives=positives, negatives=negatives
                )


# A class for loading the data, need to move the other places.
class PairwiseSamplerFromTSV(PairwiseSampler):

    pairwise_samples_path: Param[Path]
    """The path which stores the existing triplets"""

    def pairwise_iter(self) -> SerializableIterator[PairwiseRecord, Any]:
        def iter() -> Iterator[PairwiseSample]:
            for triplet in read_tsv(self.pairwise_samples_path):
                q_id, pos_id, pos_score, neg_id, neg_score = triplet
                yield PairwiseRecord(
                    Record(IDItem(q_id)),
                    Record(IDItem(pos_id), ScoredItem(pos_score)),
                    Record(IDItem(neg_id), ScoredItem(neg_score)),
                )

        return SkippingIterator(iter)


# --- Tasks for hard negatives


class ModelBasedHardNegativeSampler(Task, Sampler):
    """Retriever-based hard negative sampler"""

    dataset: Param[Adhoc]
    """The dataset which contains the topics and assessments"""

    retriever: Param[Retriever]
    """The retriever to score of the document wrt the query"""

    hard_negative_samples: Annotated[Path, pathgenerator("hard_negatives.tsv")]
    """Path to store the generated hard negatives"""

    def task_outputs(self, dep) -> PairwiseSampleDataset:
        """return a iterator of PairwiseSample"""
        return dep(
            TSVPairwiseSampleDataset(
                ids=self.dataset.id,
                hard_negative_samples_path=self.hard_negative_samples,
            )
        )

    def execute(self):
        """Retrieve over the dataset and select the positive and negative
        according to the relevance score and their rank
        """
        self.logger.info("Reading topics and retrieving documents")

        # create the file
        self.hard_negative_samples.parent.mkdir(parents=True, exist_ok=True)

        # Read the assessments
        self.logger.info("Reading assessments")
        assessments = {}  # type: Dict[str, Dict[str, float]]
        for qrels in self.dataset.assessments.iter():
            doc2rel = {}
            assessments[qrels.topic_id] = doc2rel
            for qrel in qrels.assessments:
                doc2rel[qrel.doc_id] = qrel.rel
        self.logger.info("Assessment loaded")
        self.logger.info("Read assessments for %d topics", len(assessments))

        self.logger.info("Retrieving documents for each topic")
        queries = []
        for query in self.dataset.topics.iter():
            queries.append(query)

        with self.hard_negative_samples.open("wt") as fp:
            # Retrieve documents

            # count the number of queries been skipped because of no assessments
            # available
            skipped = 0
            for query in tqdm(queries):
                qassessments = assessments.get(query[IDItem].id, None)
                if not qassessments:
                    skipped += 1
                    self.logger.warning("Skipping topic %s (no assessments)", query[IDItem].id)
                    continue

                # Write all the positive documents
                positives = []
                negatives = []
                scoreddocuments: List[ScoredDocument] = self.retriever.retrieve(query)

                for rank, sd in enumerate(scoreddocuments):
                    if qassessments.get(sd.document[IDItem].id, 0) > 0:
                        # It is a positive document:
                        positives.append(sd.document[IDItem].id)
                    else:
                        # It is a negative document or
                        # don't exist in assessment
                        negatives.append(sd.document[IDItem].id)

                if not positives:
                    self.logger.debug(
                        "Skipping topic %s (no relevant documents)", query[IDItem].id
                    )
                    skipped += 1
                    continue
                if not negatives:
                    self.logger.debug(
                        "Skipping topic %s (no negative documents)", query[IDItem].id
                    )
                    skipped += 1
                    continue

                # Write the result to the file
                positive_str = " ".join(positives)
                negative_str = " ".join(negatives)
                fp.write(
                    f"{query[IDItem].id}\tpositives:\t{positive_str}\t"
                    f"negatives:\t{negative_str}"
                )

        self.logger.info("Processed %d topics (%d skipped)", len(queries), skipped)


class TeacherModelBasedHardNegativesTripletSampler(Task, Sampler):
    """Builds a teacher file for pairwise distillation losses"""

    sampler: Param[PairwiseSampler]
    """The list of exsting hard negatives which we can sample from"""

    document_store: Param[DocumentStore]
    """The document store"""

    topic_store: Param[TextStore]
    """The query_document store"""

    teacher_model: Param[Scorer]
    """The teacher model which scores the positive and negative document"""

    hard_negative_triplet: Annotated[Path, pathgenerator("triplet.tsv")]
    """The path to store the generated triplets"""

    batch_size: int
    """How many pairs of documents are been calculate in a batch"""

    def task_outputs(self, dep) -> PairwiseSampler:
        return dep(
            PairwiseSamplerFromTSV(pairwise_samples_path=self.hard_negative_triplet)
        )

    def iter_pairs_with_text(self) -> Iterator[PairwiseRecord]:
        """Add the information of the text back to the records"""
        for record in self.sampler.pairwise_iter():
            record.query.text = self.topic_store[record.query.id]
            record.positive.text = self.document_store.document_text(
                record.positive.docid
            )
            record.negative.text = self.document_store.document_text(
                record.negative.docid
            )
            yield record

    def iter_batches(self) -> Iterator[PairwiseRecords]:
        """Return the batch which contains the records"""
        while True:
            batch = PairwiseRecords()
            for _, record in zip(range(self.batch_size), self.iter_pairs_with_text()):
                batch.add(record)
            yield batch

    def execute(self):
        """Pre-calculate the score for the teacher model, and store them"""

        self.logger.info("Calculating the score for the teacher model")
        # create the file
        self.hard_negative_triplet.parent.mkdir(parents=True, exist_ok=True)

        # make the tqdm progressing wrt one record, not a batch of records
        with self.hard_negative_triplet.open("wt") as fp:
            for batch in tqdm(self.iter_batches()):

                # scores in shape: [batch_size, 2]
                self.teacher_model.eval()
                scores = self.teacher_model(batch)
                scores = scores.reshape(2, -1).T

                # write in the file
                for i, record in enumerate(batch):
                    fp.write(
                        f"{record.query.id}\t{record.positive.id}\t{scores[i,0]}"
                        f"\t{record.negative.id}\t{scores[i,1]}"
                    )
        self.logger.info("Teacher models score generating finish")



from typing import Iterator, List
from xpmir.learning.batchers import PowerAdaptativeBatcher


def add(x: int, y: int):
    return x + y


class Processor:
    def __init__(self, fn):
        self.fn = fn

    def __call__(self, batch: List[int], results: List[int], *args):
        for x in batch:
            results.append(self.fn(x, *args))


class IteratorProcessor:
    def __init__(self, fn):
        self.fn = fn
        self.calls = 0

    def __call__(self, batches: Iterator[List[int]], count: int, *args):
        self.calls += 1
        results = []
        for batch in batches:
            count -= 1
            for x in batch:
                results.append(self.fn(x, *args))
        assert count == 0
        return results


def apply(fn, data: List[int], *args):
    results: List[int] = []
    for x in data:
        results.append(fn(x, *args))
    return results


class FailingAdd:
    def __init__(self, fail: int):
        self.fail = fail

    def __call__(self, x: int, y: int):
        if x == self.fail:
            self.fail = None
            # Simulates a CUDA out-of-memory error
            raise RuntimeError("CUBLAS_STATUS_ALLOC_FAILED")
        return x + y


def test_poweradaptativebatcher():
    batcher: PowerAdaptativeBatcher = PowerAdaptativeBatcher().instance()
    worker = batcher.initialize(3)

    data = list(range(7))
    args = [4]
    expected = apply(add, data, *args)

    # Test without trouble
    observed = []
    worker.process(data, Processor(add), observed, *args)
    assert observed == expected

    # Test with OOM handling
    worker = batcher.initialize(3)
    failing_add = FailingAdd(3)
    observed = []
    worker.process(data, Processor(failing_add), observed, *args)
    assert failing_add.fail is None
    assert observed == expected

    # Test with OOM and restart
    worker = batcher.initialize(3)
    failing_add = FailingAdd(3)
    processor = IteratorProcessor(failing_add)
    observed = worker.process_withreplay(data, processor, *args)
    assert failing_add.fail is None
    assert processor.calls == 2
    assert observed == expected



from typing import (
    BinaryIO,
    Callable,
    Iterator,
    List,
    TextIO,
    TypeVar,
    Union,
    Iterable,
    Tuple,
    Type,
)
import inspect
import logging
import os
from pathlib import Path
import re
from subprocess import run
import tempfile
from experimaestro import SubmitHook, Job, Launcher
from threading import Thread
from xpmir.utils.functools import cache
from xpmir.utils.logging import easylog, EasyLogger  # noqa: F401

T = TypeVar("T")


class StreamGenerator(Thread):
    """Create a FIFO pipe (*nix only) that is fed by the provider generator"""

    def __init__(self, generator: Callable[[Union[TextIO, BinaryIO]], None], mode="wb"):
        super().__init__()
        tmpdir = tempfile.mkdtemp()
        self.mode = mode
        self.filepath = Path(os.path.join(tmpdir, "fifo.json"))
        os.mkfifo(self.filepath)
        self.generator = generator
        self.error = False

    def run(self):
        try:
            with self.filepath.open(self.mode) as out:
                try:
                    self.generator(out)
                except Exception:
                    # Just write something so the file is closed
                    if isinstance(out, TextIO):
                        out.write("")
                    else:
                        out.write(b"0")
                    raise
        except Exception:
            self.error = True
            raise

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.join()
        self.filepath.unlink()
        self.filepath.parent.rmdir()
        if self.error:
            raise AssertionError("Error with the generator")


class Handler:
    """Returns a result that depends on the type of the argument

    Example:
    ```
    handler = Handler()

    @handler()
    def trectopics(topics: TrecTopics):
        return ("-topicreader", "Trec", "-topics", topics.path)

    @handler()
    def tsvtopics(topics: ir_csv.Topics):
        return ("-topicreader", "TsvInt", "-topics", topics.path)

    command.extend(handler[topics])

    ```
    """

    def __init__(self):
        self.handlers: List[Tuple[Type, Callable]] = []
        self.defaulthandler = None

    def default(self):
        assert self.defaulthandler is None

        def annotate(method):
            self.defaulthandler = method
            return method

        return annotate

    def __call__(self):
        def annotate(method):
            spec = inspect.getfullargspec(method)
            assert len(spec.args) == 1 and spec.varargs is None

            self.handlers.append((spec.annotations[spec.args[0]], method))

        return annotate

    def __getitem__(self, key):
        try:
            handler = next(
                handler
                for cls, handler in self.handlers
                if issubclass(key.__class__, cls)
            )
        except StopIteration:
            if self.default is None:
                raise RuntimeError(
                    f"No handler for {key.__class__} and no default handler"
                )
            handler = self.defaulthandler

        return handler(key)


def foreach(iterator: Iterable[T], fn: Callable[[T], None]):
    for t in map(fn, iterator):
        pass


def batchiter(batchsize: int, iter: Iterator[T], keeppartial=True) -> Iterator[List[T]]:
    """Group items together to form of list of size `batchsize`"""
    samples = []
    for sample in iter:
        samples.append(sample)
        if len(samples) % batchsize == 0:
            yield samples
            samples = []

    # Yield last samples if keeppartial is true
    if keeppartial and len(samples) > 0:
        yield samples


@cache
def find_java_home(min_version: int = 6) -> str:
    """Find JAVA HOME"""
    paths = []

    # (1) Use environment variable
    if java_home := os.environ.get("FORCE_JAVA_HOME", None):
        return java_home

    if java_home := os.environ.get("JAVA_HOME", None):
        paths.append(Path(java_home) / "bin" / "java")

    # Try java
    paths.append("java")

    # (2) Use java -XshowSettings:properties
    for p in paths:
        try:
            p = run(
                [p, "-XshowSettings:properties", "-version"],
                check=True,
                capture_output=True,
            )

            if m := re.search(
                rb".*\n\s+java.version = (\d+)\.[\d\.]+(-\w+)?\n.*",
                p.stderr,
                re.MULTILINE,
            ):
                version = int(m[1].decode())
                if min_version <= version:
                    if m := re.search(
                        rb".*\n\s+java.home = (.*)\n.*", p.stderr, re.MULTILINE
                    ):
                        return m[1].decode()
                else:
                    logging.info(
                        "Java search (version >= %d): skipping %s", min_version, p
                    )

        except Exception:
            # silently ignore
            pass

    raise FileNotFoundError(f"Java (version >= {min_version}) not found")


class needs_java(SubmitHook):
    """Experimaestro hook that ensures that JAVA_HOME is set"""

    def __init__(self, version: int):
        self.version = version

    def spec(self):
        return self.version

    def process(self, job: Job, launcher: Launcher):
        job.environ["JAVA_HOME"] = find_java_home(self.version)


class Initializable:
    """Base class for all initializable (but just once)"""

    def initialize(self, *args, **kwargs):
        """Main initialization

        Calls :py:meth:`__initialize__` once (using :py:meth:`__initialize__`)
        """
        if not self._initialized:
            self._initialized = True
            self.__initialize__(*args, **kwargs)
        self._initialized = True

    def __init__(self):
        self._initialized = False

    def __initialize__(self, *args, **kwargs):
        """Initialize the object

        Parameters depend on the actual class
        """
        pass



import itertools
import random
from collections import defaultdict
from pathlib import Path
from typing import List, Optional

import torch
from experimaestro.notifications import TaskEnv
from datamaestro_text.data.ir import TextItem, IDItem, create_record, TopicRecord

from xpmir.learning.context import TrainerContext
from xpmir.neural.dual import DualRepresentationScorer
from xpmir.rankers import ScoredDocument
from xpmir.rankers.full import FullRetrieverRescorer
from xpmir.test.utils.utils import SampleDocumentStore


class ListWrapper(list):
    device = None

    def to(self, device):
        return self

    def __getitem__(self, item):
        return ListWrapper(list.__getitem__(self, item))


class CachedRandomScorer(DualRepresentationScorer[ListWrapper[str], ListWrapper[str]]):
    def __initialize__(self, options):
        super().__initialize__(options)
        self._cache = defaultdict(lambda: random.uniform(0, 1))

    def cache(self, query: str, document: str):
        return self._cache[(query, document)]

    def encode(self, texts: List[str]):
        return ListWrapper(texts)

    def score_pairs(
        self,
        queries: ListWrapper[str],
        documents: ListWrapper[str],
        info: Optional[TrainerContext],
    ) -> torch.Tensor:
        scores = [self.cache(q, d) for q, d in zip(queries, documents)]
        return torch.DoubleTensor(scores)

    def score_product(
        self,
        queries: ListWrapper[str],
        documents: ListWrapper[str],
        info: Optional[TrainerContext],
    ) -> torch.Tensor:
        scores = []
        for q in queries:
            scores.append([self.cache(q, d) for d in documents])

        return torch.DoubleTensor(scores)

    def merge_queries(self, queries):
        return ListWrapper(itertools.chain(*queries))


class _FullRetrieverRescorer(FullRetrieverRescorer):
    def retrieve(self, record: TopicRecord):
        scored_documents = [
            # Randomly get a score (and cache it)
            ScoredDocument(
                d, self.scorer.cache(record[TextItem].text, d[TextItem].text)
            )
            for d in self.documents
        ]
        scored_documents.sort(reverse=True)
        return scored_documents


def test_fullretrieverescorer(tmp_path: Path):
    NUM_DOCS = 7
    NUM_QUERIES = 9
    TaskEnv.instance().taskpath = tmp_path

    documents = SampleDocumentStore(num_docs=NUM_DOCS)
    scorer = CachedRandomScorer()
    retriever = _FullRetrieverRescorer(documents=documents, scorer=scorer, batchsize=20)

    _retriever = retriever.instance()
    _retriever.initialize()

    # Retrieve normally
    scoredDocuments = {}
    queries = {qid: create_record(text=f"Query {qid}") for qid in range(NUM_QUERIES)}

    # Retrieve query per query
    for qid, query in queries.items():
        scoredDocuments[qid] = _retriever.retrieve(query)

    # Retrieve with batching
    all_results = _retriever.retrieve_all(queries)

    for qid, results in all_results.items():
        expected = scoredDocuments[qid]
        results.sort(reverse=True)
        expected.sort(reverse=True)

        assert [d.document[IDItem].id for d in expected] == [
            d.document[IDItem].id for d in results
        ], "Document IDs do not match"
        assert [d.score for d in expected] == [
            d.score for d in results
        ], "Scores do not match"



from functools import cached_property

from xpmir.experiments import ExperimentHelper
from xpmir.learning.optim import TensorboardService


class LearningExperimentHelper(ExperimentHelper):
    @cached_property
    def tensorboard_service(self) -> TensorboardService:
        """Returns a tensorboard service"""
        return self.xp.add_service(TensorboardService(self.xp.resultspath / "runs"))


learning_experiment = LearningExperimentHelper.decorator
"""Wraps an experiment into an experiment where a model is learned

Provides:

1. Tensorboard service
"""



import re
from typing import Iterator, NamedTuple, Protocol, Tuple


def basictokenizer(text: str) -> Iterator[Tuple[str, int, int]]:
    for m in re.finditer(r"\w+", text):
        yield m.group(), m.start(), m.end()



import os
import pytest

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

IN_CONTINUOUS_INTEGRATION = IN_GITHUB_ACTIONS


def skip_if_ci(*args, **kwargs):
    return pytest.mark.skipif(
        IN_CONTINUOUS_INTEGRATION, reason="Test doesn't run in CI."
    )(*args, **kwargs)



"""Complements ir-metrics package"""

import builtins
from pkg_resources import parse_version
import ir_measures

if getattr(builtins, "__sphinx_build__", False):
    evaluator = None
else:
    evaluator = ir_measures.DefaultPipeline.evaluator

    # If new measures have to be defined, before they make it upstream:
    #
    # DefaultPipeline = ir_measures.providers.FallbackProvider(
    #     ir_measures.DefaultPipeline.providers + [AccuracyProvider()]
    # )
    # evaluator = DefaultPipeline.evaluator



import torch
from torch import nn


# The code below is heavily borrowed from OpenNIR


def binmat(a, b, padding=None):
    BAT, A, B = a.shape[0], a.shape[1], b.shape[1]
    a = a.reshape(BAT, A, 1)
    b = b.reshape(BAT, 1, B)
    result = a == b
    if padding is not None:
        result = result & (a != padding) & (b != padding)
    return result.float()


def cos_simmat(a, b, amask=None, bmask=None):
    BAT, A, B = a.shape[0], a.shape[1], b.shape[1]
    a_denom = a.norm(p=2, dim=2).reshape(BAT, A, 1) + 1e-9  # avoid 0div
    b_denom = b.norm(p=2, dim=2).reshape(BAT, 1, B) + 1e-9  # avoid 0div
    result = a.bmm(b.permute(0, 2, 1)) / (a_denom * b_denom)
    if amask is not None:
        result = result * amask.reshape(BAT, A, 1)
    if bmask is not None:
        result = result * bmask.reshape(BAT, 1, B)
    return result


class InteractionMatrix(nn.Module):
    def __init__(self, padding=-1):
        super().__init__()
        self.padding = padding

    def forward(self, a_embed, b_embed, a_tok, b_tok):
        def wrap_list(x):
            return x if isinstance(x, list) else [x]

        a_embed = wrap_list(a_embed)
        b_embed = wrap_list(b_embed)

        BAT, A, B = a_embed[0].shape[0], a_embed[0].shape[1], b_embed[0].shape[1]

        simmats = []

        for a_emb, b_emb in zip(a_embed, b_embed):
            if (
                a_emb.dtype is torch.long
                and len(a_emb.shape) == 2
                and b_emb.dtype is torch.long
                and len(b_emb.shape) == 2
            ):
                # binary matrix
                simmats.append(binmat(a_emb, b_emb, padding=self.padding))
            else:
                # cosine similarity matrix
                a_mask = (a_tok.reshape(BAT, A, 1) != self.padding).float()
                b_mask = (b_tok.reshape(BAT, 1, B) != self.padding).float()
                simmats.append(cos_simmat(a_emb, b_emb, a_mask, b_mask))
        return torch.stack(simmats, dim=1)



import torch
from typing import Optional


def to_device(tensor: Optional[torch.Tensor], device: torch.device):
    """Move to device if not None"""
    if tensor is not None:
        return tensor.to(device)
    return tensor



from enum import Enum
import logging
import torch
from pathlib import Path
from typing import Dict, Iterator, List, NamedTuple, Any
from experimaestro import (
    Task,
    Config,
    Param,
    pathgenerator,
    Annotated,
    tqdm,
    Meta,
)
import numpy as np
from xpmir.context import Hook, InitializationHook
from xpmir.utils.utils import EasyLogger, easylog, foreach
from xpmir.learning.devices import DEFAULT_DEVICE, Device, DeviceInformation
from xpmir.learning import Random, ModuleInitMode
from xpmir.learning.trainers import Trainer
from xpmir.learning.context import (
    StepTrainingHook,
    TrainState,
    TrainerContext,
)
from xpmir.learning.metrics import Metrics

from .optim import (
    Module,
    ModuleLoader,
    ParameterOptimizer,
    ScheduledOptimizer,
    OptimizationHook,
)
from .batchers import RecoverableOOMError

logger = easylog()


class LearnerListenerStatus(Enum):
    NO_DECISION = 0
    STOP = 1
    DONT_STOP = 2

    def update(self, other: "LearnerListenerStatus") -> "LearnerListenerStatus":
        return LearnerListenerStatus(max(self.value, other.value))


class LearnerListener(Config):
    """Hook for learner

    Performs some operations after a learning epoch"""

    id: Meta[str]
    """Unique ID to identify the listener (ignored for signature)"""

    def initialize(self, learner: "Learner", context: TrainerContext):
        self.learner = learner
        self.context = context

    def __call__(self, state: TrainState) -> LearnerListenerStatus:
        """Process and returns whether the training process should stop"""
        return LearnerListenerStatus.NO_DECISION

    def update_metrics(self, metrics: Dict[str, float]):
        """Add metrics"""
        pass

    def init_task(self, learner: "Learner", dep):
        """Returns the initialization task that loads the associated checkpoint

        :param learner: The learner object
        :param dep: The function that adds a dependency
        """
        return None


class LearnerOutput(NamedTuple):
    """The data structure for the output of a learner. It contains a dictionary
    where the key is the name of the listener and the value is the output of
    that listener"""

    listeners: Dict[str, Any]

    learned_model: ModuleLoader


class Learner(Task, EasyLogger):
    """Model Learner

    The learner task is generic, and takes two main arguments: (1) the model
    defines the model (e.g. DRMM), and (2) the trainer defines how the model
    should be trained (e.g. pointwise, pairwise, etc.)

    When submitted, it returns a dictionary based on the `listeners`
    """

    # Training
    random: Param[Random]
    """The random generator"""

    trainer: Param[Trainer]
    """Specifies how to train the model"""

    model: Param[Module]
    """Defines the model to be learned. If multiple models are used, one can use
    MultipleModel.
    """

    max_epochs: Param[int] = 1000
    """Maximum number of epochs"""

    steps_per_epoch: Param[int] = 128
    """Number of steps for one epoch (after each epoch results are logged)"""

    use_fp16: Param[bool] = False
    """Use mixed precision when training"""

    optimizers: Param[List[ParameterOptimizer]]
    """The list of parameter optimizers"""

    listeners: Param[List[LearnerListener]]
    """Listeners are in charge of handling the validation of the model, and
    saving the relevant checkpoints"""

    checkpoint_interval: Param[int] = 1
    """Number of epochs between each checkpoint"""

    logpath: Annotated[Path, pathgenerator("runs")]
    """The path to the tensorboard logs"""

    checkpointspath: Annotated[Path, pathgenerator("checkpoints")]
    """The path to the checkpoints"""

    device: Meta[Device] = DEFAULT_DEVICE
    """The device(s) to be used for the model"""

    hooks: Param[List[Hook]] = []
    """Global learning hooks

    :class:`Initialization hooks <xpmir.context.InitializationHook>` are called
    before and after the initialization of the trainer and listeners.
    """

    def __validate__(self):
        assert self.optimizers, "At least one optimizer should be defined"
        assert len(set(listener.id for listener in self.listeners)) == len(
            self.listeners
        ), "IDs of listeners should be unique"
        return super().__validate__()

    def task_outputs(self, dep) -> LearnerOutput:
        """Object returned when submitting the task"""
        return LearnerOutput(
            listeners={
                listener.id: listener.init_task(self, dep)
                for listener in self.listeners
            },
            learned_model=dep(
                ModuleLoader.C(
                    value=self.model,
                    path=self.last_checkpoint_path / TrainState.MODEL_PATH,
                )
            ),
        )

    @property
    def last_checkpoint_path(self):
        return self.checkpointspath / "last"

    def execute(self):
        self.device.execute(self.device_execute)

    def device_execute(self, device_information: DeviceInformation):
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        for handler in logger.handlers:
            handler.setLevel(logging.INFO)

        self.optimizer = ScheduledOptimizer()
        self.only_cached = False
        self.context = TrainerContext(
            device_information,
            self.logpath,
            self.checkpointspath,
            self.max_epochs,
            self.steps_per_epoch,
            self.trainer,
            self.model,
            self.optimizer,
        )

        for hook in self.hooks:
            self.context.add_hook(hook)

        # Call hooks
        foreach(
            self.context.hooks(InitializationHook),
            lambda hook: hook.before(self.context),
        )

        # Sets the random seed
        seed = self.random.state.randint((2**32) - 1)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

        # Initialize the scorer and trainer
        self.logger.info("model initialization")
        self.model.initialize(ModuleInitMode.DEFAULT.to_options(self.random.state))

        # Initialize the context and the listeners
        self.trainer.initialize(self.random.state, self.context)
        for listener in self.listeners:
            listener.initialize(self, self.context)

        self.logger.info("Moving model to device %s", device_information.device)
        self.model.to(device_information.device)
        self.trainer.to(device_information.device)
        num_training_steps = self.max_epochs * self.steps_per_epoch
        self.optimizer.initialize(
            self.optimizers,
            num_training_steps,
            self.model,
            self.use_fp16,
            hooks=[hook for hook in self.hooks if isinstance(hook, OptimizationHook)],
            trainer_context=self.context,
        )

        foreach(
            self.context.hooks(InitializationHook),
            lambda hook: hook.after(self.context),
        )

        self.logger.info("Starting to train")

        current = 0
        state = None

        with tqdm(
            total=self.max_epochs, desc=f"Training ({self.max_epochs} epochs)"
        ) as tqdm_epochs:
            for state in self.iter_train(device_information):
                # Report progress
                tqdm_epochs.update(state.epoch - current)
                current = state.epoch

                if state.epoch >= 0 and not self.only_cached:
                    message = f"epoch {state.epoch}"
                    if state.cached:
                        self.logger.debug(f"[train] [cached] {message}")
                    else:
                        self.logger.debug(f"[train] {message}")

                if state.epoch == -1:
                    continue

                if not state.cached and state.epoch % self.checkpoint_interval == 0:
                    # Save checkpoint if needed
                    self.context.save_checkpoint()
                    self.context.copy(self.last_checkpoint_path)

                # Call listeners
                decision = LearnerListenerStatus.NO_DECISION
                for listener in self.listeners:
                    # listener.__call__ returns True if we should stop
                    decision = decision.update(listener(state))

                if decision == LearnerListenerStatus.STOP:
                    self.logger.warning(
                        "stopping after epoch {epoch} ({early_stop} epochs) since "
                        "all listeners asked for it"
                    )
                    break

                # Stop if max epoch is reached
                if self.context.epoch >= self.max_epochs:
                    self.logger.warning(
                        "stopping after epoch {max_epochs} (max_epoch)".format(
                            **self.__dict__
                        )
                    )
                    break

            # End of the learning process
            if state is not None and not state.cached:
                # Set the hyper-parameters
                metrics = {}
                for listener in self.listeners:
                    listener.update_metrics(metrics)
                self.context.writer.add_hparams(getattr(self, "__tags__", {}), metrics)

    def iter_train(self, device_information) -> Iterator[TrainState]:
        """Train iteration"""
        # Try to load a checkpoint

        if self.context.load_bestcheckpoint(self.max_epochs):
            yield self.context.state

        # Get an iterator over batches
        iter = self.trainer.iter_batches()

        while True:
            # Step to the next epoch
            self.context.nextepoch()

            # Train for an epoch
            with tqdm(
                leave=False,
                total=self.steps_per_epoch,
                ncols=100,
                desc=f"Train - epoch #{self.context.epoch}",
            ) as pbar:
                # Put the model into training mode (just in case)
                self.context.state.model.train()

                # Epoch: loop over batches
                metrics = Metrics()
                for b in range(self.steps_per_epoch):
                    # Get the next batch
                    batch = next(iter)
                    self.context.nextbatch()

                    while True:
                        try:
                            # Computes the gradient, takes a step and collect metrics
                            with self.context.step(metrics):
                                # Call the hook epoch hook
                                foreach(
                                    self.context.hooks(StepTrainingHook),
                                    lambda hook: hook.before(self.context),
                                )

                                # Computes the gradient
                                with torch.autocast(
                                    device_information.device.type,
                                    enabled=self.use_fp16,
                                ):
                                    self.trainer.process_batch(batch)

                                # Update metrics and counter
                                pbar.update(1)
                                break
                        except RecoverableOOMError:
                            logger.warning(
                                "Recoverable OOM detected"
                                " - re-running the training step"
                            )
                            continue

                    foreach(
                        self.context.hooks(StepTrainingHook),
                        lambda hook: hook.after(self.context),
                    )

                # Yields the current state (after one epoch)
                yield self.context.state

                # Report metrics over the epoch
                metrics.report(
                    self.context.state.step,
                    self.context.writer,
                    "train",
                )



import click
from experimaestro.experiments.cli import experiments_cli


@click.group()
def cli():
    pass


# Add some commands
cli.add_command(experiments_cli, "run-experiment")


@cli.group()
def huggingface():
    pass


@click.argument("hf_id", type=str)
@huggingface.command()
def preload(hf_id: str):
    """Pre-load the HuggingFace model using AutoModel and AutoTokenizer"""
    from transformers import AutoModel, AutoTokenizer

    AutoModel.from_pretrained(hf_id)
    AutoTokenizer.from_pretrained(hf_id)


def main():
    cli(obj=None)


if __name__ == "__main__":
    main()



"""Index for sparse models"""

import os
from queue import Full, Empty
import torch.multiprocessing as mp
from typing import Any, Generic, TypeVar
from xpmir.utils.logging import easylog

logger = easylog()

T = TypeVar("T")


def available_cpus():
    """Returns the number of available CPU cores"""
    try:
        return len(os.sched_getaffinity(0))
    except AttributeError:
        return mp.cpu_count()


class StoppableQueue(Generic[T]):
    """Queue with a stop event flag"""

    def __init__(self, maxsize: int, stopping_event: mp.Event):
        self.queue = mp.Queue(maxsize)
        self._stopping_event = stopping_event

    def get(self, timeout=1.0):
        item = None
        while True:
            try:
                item = self.queue.get(timeout=timeout)
                break
            except Empty:
                if self._stopping_event.is_set():
                    raise

        return item

    def put(self, item: Any, timeout=1.0):
        while True:
            try:
                self.queue.put(item, timeout=timeout)
                break
            except Full:
                # Try again...
                if self._stopping_event.is_set():
                    raise

    def close(self):
        self.queue.close()

    def stop(self):
        logger.warning("Stopping the iterator")
        self._stopping_event.set()



from xpmir.learning.context import TrainerContext
from xpmir.letor.records import BaseRecords
from xpmir.rankers import LearnableScorer
from experimaestro import Param
from transformers import AutoModelForSequenceClassification, AutoTokenizer, AutoConfig
from xpmir.text.encoders import TokenizedTexts
from typing import List, Tuple
from xpmir.distributed import DistributableModel
import torch
from datamaestro_text.data.ir import TextItem


class HFCrossScorer(LearnableScorer, DistributableModel):
    """Load a cross scorer model from the huggingface"""

    hf_id: Param[str]
    """the id for the huggingface model"""

    max_length: Param[int] = None
    """the max length for the transformer model"""

    @property
    def device(self):
        return self._dummy_param.device

    def __post_init__(self):
        self._dummy_param = torch.nn.Parameter(torch.Tensor())

        self.config = AutoConfig.from_pretrained(self.hf_id)
        self.model = AutoModelForSequenceClassification.from_pretrained(
            self.hf_id, config=self.config
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.hf_id)

    def batch_tokenize(
        self,
        input_records: BaseRecords,
        maxlen=None,
        mask=False,
    ) -> TokenizedTexts:
        """Transform the text to tokens by using the tokenizer"""
        if maxlen is None:
            maxlen = self.tokenizer.model_max_length
        else:
            maxlen = min(maxlen, self.tokenizer.model_max_length)

        texts: List[Tuple[str, str]] = [
            (q[TextItem].text, d[TextItem].text)
            for q, d in zip(input_records.queries, input_records.documents)
        ]

        r = self.tokenizer(
            texts,
            max_length=maxlen,
            truncation=True,
            padding=True,
            return_tensors="pt",
            return_length=True,
            return_attention_mask=mask,
        )
        return TokenizedTexts(
            None,
            r["input_ids"].to(self.device),
            r["length"],
            r.get("attention_mask", None),
            r.get("token_type_ids", None),  # if r["token_type_ids"] else None
        )

    def forward(self, inputs: BaseRecords, info: TrainerContext = None):

        tokenized = self.batch_tokenize(inputs, maxlen=self.max_length, mask=True)
        # strange that some existing models on the huggingface don't use the token_type
        with torch.set_grad_enabled(torch.is_grad_enabled()):
            result = self.model(
                tokenized.ids, token_type_ids=tokenized.token_type_ids.to(self.device), attention_mask=tokenized.mask.to(self.device)
            ).logits  # Tensor[float] of length records size
        return result

    def distribute_models(self, update):
        self.model = update(self.model)



from typing import List, Tuple, Optional
from attr import define
from datamaestro_text.data.conversation.base import EntryType
import torch
import sys
from experimaestro import Param, Constant
from datamaestro.record import Record
from datamaestro_text.data.ir import TextItem
from datamaestro_text.data.conversation import (
    AnswerEntry,
    ConversationHistoryItem,
)
from xpmir.conversation.learning.reformulation import (
    ConversationRepresentationEncoder,
)
from xpmir.text.encoders import (
    RepresentationOutput,
    TextsRepresentationOutput,
)
from xpmir.letor.trainers.alignment import AlignmentLoss
from xpmir.neural.splade import SpladeTextEncoderV2
from xpmir.utils.logging import easylog

logger = easylog()


@define
class CoSPLADEOutput(RepresentationOutput):
    q_queries: torch.Tensor
    q_answers: torch.Tensor


class AsymetricMSEContextualizedRepresentationLoss(
    AlignmentLoss[CoSPLADEOutput, TextsRepresentationOutput]
):
    """Computes the asymetric loss for CoSPLADE"""

    version: Constant[int] = 3
    """Current version
    
    version 3: uses q_answers and not q_cosplade
    """

    def __call__(self, input: CoSPLADEOutput, target: TextsRepresentationOutput):
        # Builds up the list of tokens in the gold output
        ids = target.tokenized.ids.cpu()
        sources = []
        tokens = []
        for ix, (ids, length) in enumerate(
            zip(target.tokenized.ids, target.tokenized.lens)
        ):
            for token_id in set(ids[:length]):
                sources.append(ix)
                tokens.append(token_id)

        # Compute the loss
        delta = (
            torch.relu(target.value[sources, tokens] - input.q_answers[sources, tokens])
            ** 2
        )
        return torch.sum(delta) / input.value.numel()


class CoSPLADE(ConversationRepresentationEncoder):
    """CoSPLADE model"""

    history_size: Param[int] = 0
    """Size of history to take into account (0 for infinite)"""

    queries_encoder: Param[SpladeTextEncoderV2[List[List[str]]]]
    """Encoder for the query history (the first one being the current one)"""

    history_encoder: Param[SpladeTextEncoderV2[Tuple[str, str]]]
    """Encoder for (query, answer) pairs"""

    version: Constant[int] = 2
    """Current version"""

    reverse_queries: Param[bool] = False
    """If True, use the order q_n, q_1, ..., q_{n-1}. If False, q_n, q_{n-1}, ..., q_1

    The original CoSPLADE uses True for this parameter
    """

    def __initialize__(self, options):
        super().__initialize__(options)

        self.queries_encoder.initialize(options)
        self.history_encoder.initialize(options)

    @property
    def dimension(self):
        return self.queries_encoder.dimension

    def forward(self, records: List[Record]):
        queries: List[List[str]] = []
        query_answer_pairs: List[Tuple[str, str]] = []
        pair_origins: List[int] = []
        history_size = self.history_size or sys.maxsize

        # Process each topic record

        #: History size for normalization
        history_sizes = torch.zeros((len(records), 1))

        for ix, c_record in enumerate(records):
            # Adds q_n, q_{n-1}, ..., q_{1}
            q_history = [
                entry[TextItem].text
                for entry in c_record[ConversationHistoryItem].history
                if entry[EntryType] == EntryType.USER_QUERY
            ]

            if self.reverse_queries:
                # Adds q_n, q_1, ..., q_{n-1}
                q_history.reverse()

            queries.append([c_record[TextItem].text] + q_history)

            # List of query/answer couples
            answer: Optional[AnswerEntry] = None
            count = 0
            for item in c_record[ConversationHistoryItem].history:
                entry_type = item[EntryType]
                if entry_type == EntryType.USER_QUERY and answer is not None:
                    count += 1
                    query_answer_pairs.append((c_record[TextItem].text, answer.answer))
                    pair_origins.append(ix)
                    if len(pair_origins) >= history_size:
                        break
                elif entry_type == EntryType.SYSTEM_ANSWER:
                    if (answer := item.get(AnswerEntry)) is None:
                        logger.warning("Answer record has no answer entry")
                else:
                    # Ignore anything which is not a pair topic-response
                    answer = None

            history_sizes[ix, 0] = max(count, 1)

        # (1) encodes the queries
        q_queries = self.queries_encoder(queries).value

        # (2) encodes the past queries and answers (if any)
        q_answers = torch.zeros_like(q_queries)
        if query_answer_pairs:
            x_pairs = self.history_encoder(query_answer_pairs).value
            q_ix = torch.tensor(pair_origins, dtype=torch.long, device=q_queries.device)
            q_ix = q_ix.unsqueeze(-1).expand(x_pairs.shape)
            q_answers.scatter_add_(0, q_ix, x_pairs)

            # Normalize by number of pairs
            q_answers /= history_sizes.to(q_queries.device)

        return CoSPLADEOutput(q_queries + q_answers, q_queries, q_answers)
